from typing import Optional, List, Dict

from pydantic import Field, BaseModel, ConfigDict

from thestage.services.clients.thestage_api.dtos.docker_container_assigned_device import DockerContainerAssignedDeviceDto
from thestage.services.clients.thestage_api.dtos.docker_container_mapping import DockerContainerMappingDto
from thestage.services.clients.thestage_api.dtos.enums.container_pending_action import DockerContainerAction
from thestage.services.clients.thestage_api.dtos.enums.container_status import DockerContainerStatus
from thestage.services.clients.thestage_api.dtos.frontend_status import FrontendStatusDto
from thestage.services.clients.thestage_api.dtos.installed_service import DockerContainerInstalledServicesDto
from thestage.services.clients.thestage_api.dtos.instance_rented_response import InstanceRentedDto
from thestage.services.clients.thestage_api.dtos.project_response import ProjectDto
from thestage.services.clients.thestage_api.dtos.selfhosted_instance_response import SelfHostedInstanceDto
from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse, TheStageBasePaginatedResponse
from thestage.services.clients.thestage_api.dtos.pagination_data import PaginationData


class DockerContainerDto(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    id: Optional[int] = Field(None, alias='id')
    instance_rented_id: Optional[int] = Field(None, alias='instanceRentedId')
    selfhosted_instance_id: Optional[int] = Field(None, alias='selfhostedInstanceId')
    instance_rented: Optional[InstanceRentedDto] = Field(None, alias='instanceRented')
    selfhosted_instance: Optional[SelfHostedInstanceDto] = Field(None, alias='selfhostedInstance')

    project_id: Optional[int] = Field(None, alias='projectId')
    project: Optional[ProjectDto] = Field(None, alias='project')

    is_default: Optional[bool] = Field(None, alias='isDefault')
    system_name: Optional[str] = Field(None, alias='systemName')
    title: Optional[str] = Field(None, alias='title')
    slug: Optional[str] = Field(None, alias='slug')
    docker_image: Optional[str] = Field(None, alias='dockerImage')
    assigned_devices: Optional[DockerContainerAssignedDeviceDto] = Field(None, alias='assignedDevices')
    mappings: Optional[DockerContainerMappingDto] = Field(None, alias='mappings')

    frontend_status: Optional[FrontendStatusDto] = Field(None, alias='frontendStatus')

    pending_action: Optional[DockerContainerAction] = Field(None, alias='pendingAction')
    installed_services: Optional[DockerContainerInstalledServicesDto] = Field(None, alias='installedServices')
    created_at: Optional[str] = Field(None, alias='createdAt')
    updated_at: Optional[str] = Field(None, alias='updatedAt')


class DockerContainerPaginated(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    entities: List[DockerContainerDto] = Field(default_factory=list, alias='entities')
    current_page: Optional[int] = Field(None, alias='currentPage')
    last_page: Optional[bool] = Field(None, alias='lastPage')
    total_pages: Optional[int] = Field(None, alias='totalPages')
    pagination_data: Optional[PaginationData] = Field(None, alias='paginationData')

#
# class DockerContainerListResponse(TheStageBasePaginatedResponse):
#     paginated_list: Optional[DockerContainerPaginated] = Field(None, alias='paginatedList')


class DockerContainerViewResponse(TheStageBaseResponse):
    docker_container: Optional[DockerContainerDto] = Field(None, alias='dockerContainer')


class ContainerBusinessStatusMapperResponse(TheStageBasePaginatedResponse):
    model_config = ConfigDict(use_enum_values=True)

    docker_container_status_map: Dict[str, str] = Field(default=dict, alias='dockerContainerStatusMap')



import typer

import json
import os
from json import JSONDecodeError
from pathlib import Path
from typing import Optional, Dict, Any

from thestage.exceptions.file_system_exception import FileSystemException
from thestage.services.core_files.config_entity import ConfigEntity
from thestage.helpers.ssh_util import parse_private_key
from thestage.services.connect.dto.remote_server_config import RemoteServerConfig
from thestage.services.filesystem_service import FileSystemServiceCore
from thestage.services.project.dto.project_config import ProjectConfig
from thestage.config import THESTAGE_CONFIG_DIR, THESTAGE_CONFIG_FILE, THESTAGE_AUTH_TOKEN, THESTAGE_API_URL

class ConfigProvider():
    # path to current physical subject we work with (project / etc) - might be passed explicitly
    _local_path: Optional[Path] = None
    _local_config_path: Optional[Path] = None
    _global_config_path: Optional[Path] = None
    _global_config_file: Optional[Path] = None
    _file_system_service = FileSystemServiceCore


    def __init__(
            self,
            local_path: Optional[str] = None,
    ):
        self._file_system_service = FileSystemServiceCore()
        if local_path:
            self._local_path = self._file_system_service.get_path(directory=local_path, auto_create=False)

        config_folder_name = self._file_system_service.get_path(f"{THESTAGE_CONFIG_DIR}", False)
        self._local_config_path = self._local_path.joinpath(config_folder_name)

        home_dir = self._file_system_service.get_home_path()
        self._global_config_path = home_dir.joinpath(config_folder_name)

        if self._global_config_path:
            if not self._global_config_path.exists():
                self._file_system_service.create_if_not_exists_dir(self._global_config_path)

        self._global_config_file = self._global_config_path.joinpath(
            self._file_system_service.get_path(f"{THESTAGE_CONFIG_FILE}", False)
        )
        if not self._global_config_file.exists():
            self._file_system_service.create_if_not_exists_file(self._global_config_file)


    def get_full_config(self) -> ConfigEntity:
        config_values = {}

        config_from_env = {}
        config_from_env['main'] = {}
        if THESTAGE_AUTH_TOKEN:
            config_from_env['main']['thestage_auth_token'] = THESTAGE_AUTH_TOKEN
        if THESTAGE_API_URL:
            config_from_env['main']['thestage_api_url'] = THESTAGE_API_URL

        if config_from_env:
            self.__update_config_values_dict(values_to_update=config_values, new_values=config_from_env)

        config_from_file = self.__read_config_file(self._global_config_file)
        if config_from_file:
            self.__update_config_values_dict(values_to_update=config_values, new_values=config_from_file)

        config = ConfigEntity.model_validate(config_values)

        if self._local_path:
            config.runtime.working_directory = str(self._local_path)
        if self._global_config_path and not config.runtime.config_global_path:
            config.runtime.config_global_path = str(self._global_config_path)

        return config


    def save_project_config(self, project_config: ProjectConfig):
        project_data_dirpath = self.__get_project_config_path()
        if not project_data_dirpath.exists():
            self._file_system_service.create_if_not_exists_dir(project_data_dirpath)

        project_data_filepath = self.__get_project_config_path(with_file=True)
        if not project_data_filepath.exists():
            self._file_system_service.create_if_not_exists_file(project_data_filepath)

        project_config_path = self.__get_project_config_path(with_file=True)
        self.__save_config_file(data=project_config.model_dump(), file_path=project_config_path)


    def save_project_deploy_ssh_key(self, deploy_ssh_key: str, project_slug: str, project_id: int) -> str:
        deploy_key_dirpath = self._global_config_path.joinpath('project_deploy_keys')
        self._file_system_service.create_if_not_exists_dir(deploy_key_dirpath)

        deploy_key_filepath = deploy_key_dirpath.joinpath(f'project_deploy_key_{project_id}_{project_slug}')
        self._file_system_service.create_if_not_exists_file(deploy_key_filepath)

        text_file = open(deploy_key_filepath, "w")
        text_file.write(deploy_ssh_key)
        text_file.close()
        os.chmod(deploy_key_filepath, 0o600)

        return str(deploy_key_filepath)


    def read_project_config(self) -> Optional[ProjectConfig]:
        project_data_dirpath = self.__get_project_config_path()
        if not project_data_dirpath.exists():
            return None

        project_data_filepath = self.__get_project_config_path(with_file=True)
        if not project_data_filepath.exists():
            return None

        config_data = self.__read_config_file(project_data_filepath) if project_data_filepath and project_data_filepath.exists() else {}
        return ProjectConfig.model_validate(config_data)


    def __get_project_config_path(self, with_file: bool = False) -> Path:
        if with_file:
            return self._local_config_path.joinpath('project.json')
        else:
            return self._local_config_path


    def __get_remote_server_config_path(self) -> Path:
        return self._global_config_path.joinpath('remote_server_config.json')


    def update_remote_server_config_entry(self, ip_address: str, ssh_key_path: Optional[Path]):
        # double checking if key is parseable (fails if key is encrypted)
        # see remote_server_service for similar check
        if ssh_key_path:
            pkey = parse_private_key(str(ssh_key_path))

            if pkey is None:
                typer.echo("Could not identify provided private key (expected RSA, ECDSA, ED25519)")
                raise typer.Exit(1)

        config = self.__read_remote_server_config()
        remote_server_config_filepath = self.__get_remote_server_config_path()
        if config:
            if not config.ip_address_to_ssh_key_map:
                config.ip_address_to_ssh_key_map = {}

            existing_path = config.ip_address_to_ssh_key_map.get(ip_address)
            if ssh_key_path and existing_path != str(ssh_key_path):
                config.ip_address_to_ssh_key_map.update({ip_address: str(ssh_key_path)})
                typer.echo(f'Updated ssh key for {ip_address}: {ssh_key_path}')

            if not ssh_key_path and existing_path:
                typer.echo(f"Private key at path {existing_path} was not found")
                config.ip_address_to_ssh_key_map.pop(ip_address, None)
        else:
            self._file_system_service.create_if_not_exists_file(remote_server_config_filepath)
            config = RemoteServerConfig(ip_address_to_ssh_key_map={ip_address: str(ssh_key_path)})

        self.__save_config_file(data=config.model_dump(), file_path=remote_server_config_filepath)

        return str(remote_server_config_filepath)


    def __read_remote_server_config(self) -> Optional[RemoteServerConfig]:
        config_filepath = self.__get_remote_server_config_path()
        if not config_filepath.is_file():
            return None

        config_data = self.__read_config_file(config_filepath) if config_filepath and config_filepath.exists() else {}
        return RemoteServerConfig.model_validate(config_data)


    def get_valid_private_key_path_by_ip_address(self, ip_address: str) -> Optional[str]:
        remote_server_config = self.__read_remote_server_config()
        if remote_server_config and remote_server_config.ip_address_to_ssh_key_map:
            private_key_path = remote_server_config.ip_address_to_ssh_key_map.get(ip_address)
            if private_key_path and Path(private_key_path).is_file():
                return private_key_path
            elif private_key_path:
                self.update_remote_server_config_entry(ip_address, None)
        return None


    @staticmethod
    def __update_config_values_dict(values_to_update: Dict, new_values: Dict):
        if 'main' in new_values:
            if 'main' in values_to_update:
                values_to_update['main'].update(new_values['main'])
            else:
                values_to_update['main'] = new_values['main']


    def __read_config_file(self, path: Path) -> Dict[str, Any]:
        result = {}
        try:
            if path and path.exists():
                with path.open("r") as file:
                    try:
                        if os.stat(path).st_size != 0:
                            result = json.load(file)
                    except JSONDecodeError:
                        pass
        except OSError:
            raise FileSystemException(f"Could not open config file: {path}")
        return result


    @staticmethod
    def __save_config_file(data: Dict, file_path: Path):
        with open(file_path, 'w') as configfile:
            json.dump(data, configfile, indent=1)


    def save_config(self, config: ConfigEntity):
        data: Dict[str, Any] = self.__read_config_file(self._global_config_file)
        data.update(config.model_dump(exclude_none=True, by_alias=True, exclude={'runtime', 'RUNTIME', 'daemon', 'DAEMON'}))
        self.__save_config_file(data=data, file_path=self._global_config_file)


    def clear_config(self, ):
        if self._global_config_path and self._global_config_path.exists():
            self._file_system_service.remove_folder(str(self._global_config_path))

        os.unsetenv('THESTAGE_CONFIG_DIR')
        os.unsetenv('THESTAGE_CONFIG_FILE')
        os.unsetenv('THESTAGE_CLI_ENV')
        os.unsetenv('THESTAGE_API_URL')
        os.unsetenv('THESTAGE_LOG_FILE')
        os.unsetenv('THESTAGE_AUTH_TOKEN')



from thestage.exceptions.base_exception import BaseAbstractException


class HttpClientException(BaseAbstractException):
    _status_code: int = 0

    def __init__(self, message: str, status_code: int, ):
        super(HttpClientException, self).__init__(message=message)
        self._status_code = status_code

    def get_status_code(self) -> int:
        return self._status_code



from typing import Optional

from thestage.services.connect.connect_service import ConnectService
from thestage.services.filesystem_service import FileSystemServiceCore
from thestage.services.logging.logging_service import LoggingService
from thestage.services.project.project_service import ProjectService
from thestage.services.remote_server_service import RemoteServerService
from thestage.services.container.container_service import ContainerService
from thestage.services.instance.instance_service import InstanceService
from thestage.services.app_config_service import AppConfigService
from thestage.services.clients.git.git_client import GitLocalClient
from thestage.services.clients.thestage_api.api_client import TheStageApiClient
from thestage.services.config_provider.config_provider import ConfigProvider
from thestage.services.validation_service import ValidationService


class ServiceFactory:
    __config_provider: Optional[ConfigProvider] = None
    __thestage_api_client: Optional[TheStageApiClient] = None
    __git_local_client: Optional[GitLocalClient] = None
    __file_system_service: Optional[FileSystemServiceCore] = None

    def __init__(
            self,
            config_provider: Optional[ConfigProvider] = None,
    ):
        self.__config_provider = config_provider

    def get_config_provider(self) -> ConfigProvider:
        return self.__config_provider

    def get_validation_service(self, config_provider: Optional[ConfigProvider] = None,) -> ValidationService:
        return ValidationService(
                thestage_api_client=self.get_thestage_api_client(),
                config_provider=config_provider if config_provider else self.__config_provider,
            )

    def get_instance_service(self, config_provider: Optional[ConfigProvider] = None) -> InstanceService:
        return InstanceService(
                thestage_api_client=self.get_thestage_api_client(),
                config_provider=config_provider if config_provider else self.__config_provider,
                remote_server_service=self.get_remote_server_service(),
            )

    def get_container_service(self, config_provider: Optional[ConfigProvider] = None) -> ContainerService:
        return ContainerService(
                thestage_api_client=self.get_thestage_api_client(),
                config_provider=config_provider if config_provider else self.__config_provider,
                remote_server_service=self.get_remote_server_service(),
                file_system_service=self.get_file_system_service(),
            )

    def get_connect_service(self, config_provider: Optional[ConfigProvider] = None) -> ConnectService:
        return ConnectService(
            config_provider=config_provider if config_provider else self.__config_provider,
            thestage_api_client=self.get_thestage_api_client(),
            instance_service=self.get_instance_service(),
            container_service=self.get_container_service(),
            logging_service=self.get_logging_service(),
        )

    def get_project_service(self, config_provider: Optional[ConfigProvider] = None) -> ProjectService:
        return ProjectService(
            thestage_api_client=self.get_thestage_api_client(),
            config_provider=config_provider if config_provider else self.__config_provider,
            remote_server_service=self.get_remote_server_service(),
            file_system_service=self.get_file_system_service(),
            git_local_client=self.get_git_local_client(),
        )

    def get_remote_server_service(self) -> RemoteServerService:
        return RemoteServerService(
            config_provider=self.get_config_provider(),
            file_system_service=self.get_file_system_service(),
        )

    def get_thestage_api_client(self) -> TheStageApiClient:
        if not self.__thestage_api_client:
            self.__thestage_api_client = TheStageApiClient(api_url=self.__config_provider.get_full_config().main.thestage_api_url)
        return self.__thestage_api_client

    def get_git_local_client(self):
        if not self.__git_local_client:
            self.__git_local_client = GitLocalClient(file_system_service=self.get_file_system_service())
        return self.__git_local_client

    def get_file_system_service(self) -> FileSystemServiceCore:
        if not self.__file_system_service:
            self.__file_system_service = FileSystemServiceCore()
        return self.__file_system_service

    def get_app_config_service(self, config_provider: Optional[ConfigProvider] = None,) -> AppConfigService:
        return AppConfigService(
            validation_service=self.get_validation_service(config_provider),
            config_provider=config_provider if config_provider else self.__config_provider,
        )

    def get_logging_service(self, config_provider: Optional[ConfigProvider] = None) -> LoggingService:
        return LoggingService(
            thestage_api_client=self.get_thestage_api_client(),
            config_provider=config_provider if config_provider else self.__config_provider,
        )




from typing import Optional

from pydantic import BaseModel, Field


class PaginationData(BaseModel):
    current_page: Optional[int] = Field(None, alias='currentPage')
    limit: Optional[int] = Field(None, alias='limit')
    total_pages: Optional[int] = Field(None, alias='totalPages')
    objects_count: Optional[int] = Field(None, alias='objectsCount')



from typing import Optional, List, Dict

from pydantic import Field, BaseModel, ConfigDict


class DockerContainerMappingDto(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    directory_mappings: Dict[str, str] = Field(default_factory=dict, alias='directoryMappings')
    port_mappings: Dict[str, str] = Field(default_factory=dict, alias='portMappings')



from typing import Optional
import typer

from thestage.i18n.translation import __
from thestage.services.clients.thestage_api.core.http_client_exception import HttpClientException
from thestage.services.clients.thestage_api.dtos.enums.container_status import DockerContainerStatus
from thestage.services.clients.thestage_api.dtos.enums.selfhosted_status import SelfhostedBusinessStatus
from thestage.services.clients.thestage_api.dtos.enums.instance_rented_status import InstanceRentedBusinessStatus
from thestage.services.abstract_service import AbstractService
from thestage.helpers.error_handler import error_handler
from thestage.services.clients.thestage_api.api_client import TheStageApiClient
from thestage.services.clients.thestage_api.dtos.enums.task_status import TaskStatus
from thestage.services.clients.thestage_api.dtos.instance_rented_response import InstanceRentedDto
from thestage.services.config_provider.config_provider import ConfigProvider
from thestage.services.container.container_service import ContainerService
from thestage.services.instance.instance_service import InstanceService
from thestage.services.logging.logging_service import LoggingService
from thestage.services.task.dto.task_dto import TaskDto


class ConnectService(AbstractService):
    __thestage_api_client: TheStageApiClient = None
    __instance_service: InstanceService = None
    __container_service: ContainerService = None
    __logging_service: LoggingService = None

    def __init__(
            self,
            config_provider: ConfigProvider,
            thestage_api_client: TheStageApiClient,
            instance_service: InstanceService,
            container_service: ContainerService,
            logging_service: LoggingService,
    ):
        super(ConnectService, self).__init__(
            config_provider=config_provider,
        )
        self.__thestage_api_client = thestage_api_client
        self.__instance_service = instance_service
        self.__container_service = container_service
        self.__logging_service = logging_service


    @error_handler()
    def connect_to_entity(
            self,
            uid: str,
            username: Optional[str],
            private_key_path: Optional[str],
    ):
        config = self._config_provider.get_full_config()

        # TODO make a single separate method that will return all needed info
        try:
            instance_selfhosted = self.__thestage_api_client.get_selfhosted_item(token=config.main.thestage_auth_token, instance_slug=uid)
        except HttpClientException as e:
            instance_selfhosted = None

        try:
            instance_rented = self.__thestage_api_client.get_rented_item(token=config.main.thestage_auth_token, instance_slug=uid)
        except HttpClientException as e:
            instance_rented = None

        try:
            container = self.__thestage_api_client.get_container(token=config.main.thestage_auth_token, container_slug=uid, )
        except Exception as e:
            container = None

        task: Optional[TaskDto] = None
        if uid.isdigit():
            try:
                task_view_response = self.__thestage_api_client.get_task(token=config.main.thestage_auth_token, task_id=int(uid))
            except HttpClientException as e:
                task_view_response = None
            if task_view_response and task_view_response.task:
                task = task_view_response.task

        rented_exists = int(instance_rented is not None and instance_rented.frontend_status.status_key == InstanceRentedBusinessStatus.ONLINE)
        selfhosted_exists = int(instance_selfhosted is not None)
        container_exists = int(container is not None)
        task_exists = int(task is not None)

        rented_presence = int(rented_exists and instance_rented.frontend_status.status_key == InstanceRentedBusinessStatus.ONLINE)
        selfhosted_presence = int(selfhosted_exists and instance_selfhosted.frontend_status.status_key == SelfhostedBusinessStatus.RUNNING)
        container_presence = int(container_exists and (container.frontend_status.status_key == DockerContainerStatus.RUNNING or container.frontend_status.status_key == DockerContainerStatus.BUSY))
        task_presence = int(task_exists and task.frontend_status.status_key in [TaskStatus.RUNNING, TaskStatus.SCHEDULED])

        if rented_exists:
            typer.echo(__("Found a rented instance with the provided UID in status: '%rented_status%'", {"rented_status": instance_rented.frontend_status.status_translation}))

        if selfhosted_exists:
            typer.echo(__("Found a self-hosted instance with the provided UID in status: '%selfhosted_status%'", {"selfhosted_status": instance_selfhosted.frontend_status.status_translation}))

        if container_exists:
            typer.echo(__("Found a docker container with the provided UID in status: '%container_status%'", {"container_status": container.frontend_status.status_translation}))

        if task_exists:
            typer.echo(__("Found a task with the provided ID in status: '%task_status%'", {"task_status": task.frontend_status.status_translation}))

        if (rented_presence + selfhosted_presence + container_presence + task_presence) > 1:
            typer.echo("Provided identifier caused ambiguity")
            typer.echo("Consider running a dedicated command to connect to the entity you need")
            raise typer.Exit(code=1)

        if (rented_presence + selfhosted_presence + container_presence + task_presence) == 0:
            typer.echo("There is nothing to connect to with the provided identifier")
            raise typer.Exit(code=1)

        if rented_presence:
            typer.echo("Connecting to rented instance...")
            self.__instance_service.connect_to_rented_instance(
                instance_rented_slug=uid,
                config=config,
                input_ssh_key_path=private_key_path
            )

        if container_presence:
            typer.echo("Connecting to docker container...")
            self.__container_service.connect_to_container(
                config=config,
                container_uid=uid,
                username=username,
                input_ssh_key_path=private_key_path
            )

        if selfhosted_presence:
            typer.echo("Connecting to self-hosted instance...")

            self.__instance_service.connect_to_selfhosted_instance(
                config=config,
                selfhosted_instance_slug=uid,
                username=username,
                input_ssh_key_path=private_key_path
            )

        if task_presence:
            typer.echo(__("Connecting to task..."))
            self.__logging_service.stream_task_logs_with_controls(
                config=config,
                task_id=int(uid),
            )


    @error_handler()
    def upload_ssh_key(self, public_key_contents: str, instance_slug: Optional[str]):
        config = self._config_provider.get_full_config()

        instance_rented: Optional[InstanceRentedDto] = None
        if instance_slug:
            try:
                instance_rented = self.__thestage_api_client.get_rented_item(token=config.main.thestage_auth_token, instance_slug=instance_slug)
            except HttpClientException as e:
                instance_rented = None

            # if no instances found - exit 1
            if instance_rented is None:
                typer.echo(f"No rented instance found with matching unique ID '{instance_slug}'")
                raise typer.Exit(1)

        note_to_send: Optional[str] = None

        is_user_already_has_key_response = self.__thestage_api_client.is_user_has_ssh_public_key(
            token=config.main.thestage_auth_token,
            public_key=public_key_contents
        )

        ssh_key_pair_id = is_user_already_has_key_response.sshKeyPairId
        is_adding_key_to_user = not is_user_already_has_key_response.isUserHasPublicKey

        if is_adding_key_to_user and not note_to_send:
            note_to_send: str = typer.prompt(
                text=__('SSH key will be added to your profile. Please provide a title for this key'),
                show_choices=False,
                type=str,
                show_default=False,
            )

        if not is_adding_key_to_user and not instance_rented:
            typer.echo("Key already exists on your profile")

        if is_adding_key_to_user:
            add_ssh_key_to_user_response = self.__thestage_api_client.add_public_ssh_key_to_user(
                token=config.main.thestage_auth_token,
                public_key=public_key_contents,
                note=note_to_send
            )
            typer.echo(f"Public key '{note_to_send}' added to your profile")
            ssh_key_pair_id = add_ssh_key_to_user_response.sshKeyPairId

        if instance_rented:
            self.__thestage_api_client.add_public_ssh_key_to_instance_rented(
                token=config.main.thestage_auth_token,
                instance_rented_id=instance_rented.id,
                ssh_key_pair_id=ssh_key_pair_id
            )

            if instance_rented.frontend_status.status_key != InstanceRentedBusinessStatus.ONLINE:
                typer.echo(f"Rented instance '{instance_rented.slug}' status is '{instance_rented.frontend_status.status_translation}'. Key will be added as soon as it is back online.")
            else:
                typer.echo(f"Public key added to rented instance '{instance_rented.slug}'")



def main():
    try:
        import warnings
        warnings.filterwarnings(action='ignore', module='.*paramiko.*')

        from . import __app_name__

        from thestage.controllers import base_controller, container_controller, instance_controller, project_controller, \
            config_controller

        base_controller.app.add_typer(container_controller.app, name="container")
        base_controller.app.add_typer(instance_controller.app, name="instance")
        base_controller.app.add_typer(project_controller.app, name="project")
        base_controller.app.add_typer(config_controller.app, name="config")

        import thestage.config
        base_controller.app(prog_name=__app_name__)
    except KeyboardInterrupt:
        print('THESTAGE: Keyboard Interrupt')



import json
from abc import ABC
from typing import Optional

import requests

from thestage.config import THESTAGE_API_URL
from thestage.exceptions.auth_exception import AuthException
from thestage.services.clients.thestage_api.core.http_client_exception import HttpClientException


class TheStageApiClientAbstract(ABC):
    def __init__(self, timeout: int, url: Optional[str] = None):
        self.__timeout = timeout
        self.__api_url = url

    def _get_host(self, ) -> str:
        return self.__api_url or THESTAGE_API_URL

    def _request(
            self,
            method: str,
            url: str,
            data: dict = None,
            query_params: dict = None,
            headers: Optional[dict] = None,
            token: Optional[str] = None,
    ):
        if not data:
            data = {}

        host = self._get_host()
        url = f'{host}{url}'

        request_headers = {
            'Content-Type': 'application/json',
        }

        if token:
            request_headers['Authorization'] = f"Bearer {token}"

        if headers:
            request_headers.update(headers)
        response = requests.request(
            method=method,
            url=url,
            json=data,
            params=query_params,
            headers=request_headers,
            timeout=self.__timeout,

        )
        return self._parse_api_response(response)

    @staticmethod
    def _parse_api_response(raw_response):
        content_type = raw_response.headers.get('content-type')
        message_error = None
        if content_type == 'application/json':
            try:
                result = raw_response.json()
                message_error = result.get('message', None)
            except json.JSONDecodeError:
                raise HttpClientException(
                    message=f"Failed to parse server response",
                    status_code=raw_response.status_code,
                )
        else:
            result = raw_response.content.text()

        if raw_response.status_code == 401:
            raise AuthException(
                message=f"Unauthorized",
            )
        elif raw_response.status_code == 403:
            raise HttpClientException(
                message=f"{message_error if message_error else 'Forbidden'} ({raw_response.status_code})",
                status_code=raw_response.status_code,
            )
        elif raw_response.status_code >= 400:
            raise HttpClientException(
                message=f"{message_error if message_error else 'Request error'} ({raw_response.status_code})",
                status_code=raw_response.status_code,
            )
        elif raw_response.status_code < 200 or raw_response.status_code > 300:
            raise HttpClientException(
                message=f"{message_error if message_error else 'Request error'} ({raw_response.status_code})",
                status_code=raw_response.status_code,
            )

        return result



from pydantic import Field, ConfigDict

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse


class DeployInferenceModelToSagemakerResponse(TheStageBaseResponse):
    model_config = ConfigDict(use_enum_values=True)
    modelId: str = Field(None, alias='modelId')
    ecrImageUrl: str = Field(None, alias='ecrImageUrl')
    s3ArtifactsUrl: str = Field(None, alias='s3ArtifactsUrl')





from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class DockerContainerEntity(BaseModel):

    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
    )

    # id: Optional[int] = Field(None, alias='id')
    slug: Optional[str] = Field(None, alias='UNIQUE ID')
    status: Optional[str] = Field(None, alias='STATUS')
    title: Optional[str] = Field(None, alias='TITLE')
    instance_type: Optional[str] = Field(None, alias='INSTANCE TYPE')
    instance_slug: Optional[str] = Field(None, alias='INSTANCE UNIQUE ID')
    # instance_rented_slug: Optional[str] = Field(None, alias='INSTANCE UNIQUE ID')
    # selfhosted_rented_slug: Optional[str] = Field(None, alias='SELF-HOSTED UNIQUE ID')
    # system_name: Optional[str] = Field(None, alias='SYSTEM NAME')
    # docker_image: Optional[str] = Field(None, alias='DOCKER IMAGE')



import os

import click
import typer
from thestage.services.core_files.config_entity import ConfigEntity
from thestage.i18n.translation import __
from thestage.entities.enums.yes_no_response import YesOrNoResponse
from thestage.services.config_provider.config_provider import ConfigProvider
from thestage.services.validation_service import ValidationService


class AppConfigService:

    __validation_service: ValidationService
    __config_provider: ConfigProvider

    def __init__(
            self,
            validation_service: ValidationService,
            config_provider: ConfigProvider,
    ):
        self.__validation_service = validation_service
        self.__config_provider = config_provider

    def app_change_token(
            self,
            config: ConfigEntity,
            token: str,
    ):
        if config.main.thestage_auth_token:
            response: YesOrNoResponse = typer.prompt(
                text=__('Do you want to change current token?'),
                show_choices=True,
                default=YesOrNoResponse.YES.value,
                type=click.Choice([r.value for r in YesOrNoResponse]),
                show_default=True,
            )
            if response == YesOrNoResponse.NO:
                raise typer.Exit(0)

        config.main.thestage_auth_token = token

        self.__validation_service.check_token(config=config)
        self.__config_provider.save_config(config=config)

    @staticmethod
    def app_remove_env():
        os.unsetenv('THESTAGE_CONFIG_FILE')
        os.unsetenv('THESTAGE_API_URL')
        os.unsetenv('THESTAGE_LOG_FILE')
        os.unsetenv('THESTAGE_AUTH_TOKEN')



# from thestage.exceptions.base_exception import BaseAbstractException
#
#
# class HttpClientException(BaseAbstractException):
#     _status_code: int = 0
#
#     def __init__(self, message: str, status_code: int, ):
#         super(HttpClientException, self).__init__(message=message)
#         self._status_code = status_code
#
#     def get_status_code(self) -> int:
#         return self._status_code



from typing import Optional

from pydantic import BaseModel, Field

# saved to file
class MainConfigEntity(BaseModel):
    thestage_auth_token: Optional[str] = Field(None, alias='thestage_auth_token')
    thestage_api_url: Optional[str] = Field(None, alias='thestage_api_url')


class DaemonConfigEntity(BaseModel):
    daemon_token: Optional[str] = Field(None, alias='daemon_token')
    backend_api_url: Optional[str] = Field(None, alias='backend_api_url')

# not saved to file
class RuntimeConfigEntity(BaseModel):
    working_directory: Optional[str] = Field(None, alias='working_directory')
    config_global_path: Optional[str] = Field(None, alias='config_global_path')


class ConfigEntity(BaseModel):
    main: MainConfigEntity = Field(default_factory=MainConfigEntity, alias='main')
    runtime: RuntimeConfigEntity = Field(default_factory=RuntimeConfigEntity, alias="runtime") # TODO merge with main
    daemon: DaemonConfigEntity = Field(default_factory=DaemonConfigEntity, alias="daemon") # TODO this should not be in core package
    start_on_daemon: bool = Field(False, alias='start_on_daemon') # TODO this should not be in core package



from pathlib import Path
from typing import Optional, List

from thestage.services.clients.thestage_api.dtos.enums.instance_rented_status import InstanceRentedBusinessStatus

from thestage.entities.rented_instance import RentedInstanceEntity
from thestage.entities.self_hosted_instance import SelfHostedInstanceEntity
from thestage.helpers.logger.app_logger import app_logger
from thestage.services.clients.thestage_api.dtos.enums.selfhosted_status import SelfhostedBusinessStatus
from thestage.services.instance.mapper.instance_mapper import InstanceMapper
from thestage.services.instance.mapper.selfhosted_mapper import SelfHostedMapper
from thestage.services.instance.instance_service import InstanceService
from thestage.i18n.translation import __
from thestage.controllers.utils_controller import \
    validate_config_and_get_service_factory

import typer


app = typer.Typer(no_args_is_help=True, help=__("Manage server instances"))

rented = typer.Typer(no_args_is_help=True, help=__("Manage rented server instances"))
self_hosted = typer.Typer(no_args_is_help=True, help=__("Manage self-hosted instances"))

app.add_typer(rented, name="rented")
app.add_typer(self_hosted, name="self-hosted")


@rented.command(name="ls", help=__("List rented server instances"))
def rented_list(
        row: int = typer.Option(
            5,
            '--row',
            '-r',
            help=__("Set number of rows displayed per page"),
            is_eager=False,
        ),
        page: int = typer.Option(
            1,
            '--page',
            '-p',
            help=__("Set starting page for displaying output"),
            is_eager=False,
        ),
        statuses: List[str] = typer.Option(
            None,
            '--status',
            '-s',
            help=__("Filter by status, use --status all to list all rented server instances"),
            is_eager=False,
        ),
):
    """
        Lists rented server instances
    """
    service_factory = validate_config_and_get_service_factory()
    config = service_factory.get_config_provider().get_full_config()

    instance_service: InstanceService = service_factory.get_instance_service()
    instance_rented_status_map = service_factory.get_thestage_api_client().get_rented_business_status_map(config.main.thestage_auth_token)

    if not statuses:
        statuses = ({key: instance_rented_status_map[key] for key in [
            InstanceRentedBusinessStatus.ONLINE,
            InstanceRentedBusinessStatus.CREATING,
            InstanceRentedBusinessStatus.TERMINATING,
            InstanceRentedBusinessStatus.REBOOTING,
        ]}).values()

    if "all" in statuses:
        statuses = instance_rented_status_map.values()

    for input_status_item in statuses:
        if input_status_item not in instance_rented_status_map.values():
            typer.echo(__("'%invalid_status%' is not one of %valid_statuses%", {
                'invalid_status': input_status_item,
                'valid_statuses': str(list(instance_rented_status_map.values()))
            }))
            raise typer.Exit(1)

    typer.echo(__(
        "Listing rented server instances with the following statuses: %statuses%, to view all rented server instances, use --status all",
        placeholders={
            'statuses': ', '.join([status_item for status_item in statuses])
        }))

    backend_statuses: List[str] = [key for key, value in instance_rented_status_map.items() if value in statuses]

    instance_service.print(
        func_get_data=instance_service.get_rented_list,
        func_special_params={
            'statuses': backend_statuses,
        },
        mapper=InstanceMapper(),
        config=config,
        headers=list(map(lambda x: x.alias, RentedInstanceEntity.model_fields.values())),
        row=row,
        page=page,
        show_index="never",
    )

    typer.echo(__("Rented server instances listing complete"))
    raise typer.Exit(0)


@rented.command(name="connect", no_args_is_help=True, help=__("Connect to rented server instance"))
def instance_connect(
        instance_uid: Optional[str] = typer.Argument(
            help=__("Rented server instance unique ID"),
        ),
        private_ssh_key_path: str = typer.Option(
            None,
            "--private-key-path",
            "-pk",
            help=__("Path to private key that will be accepted by remote server (optional)"),
            is_eager=False,
        )
):
    """
        Connects to rented server instance
    """
    app_logger.info(f'Connect to rented server instance')

    if not instance_uid:
        typer.echo(__('Rented server instance unique ID is required'))
        raise typer.Exit(1)

    if private_ssh_key_path and not Path(private_ssh_key_path).is_file():
        typer.echo(f'No file found at provided path {private_ssh_key_path}')
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    config = service_factory.get_config_provider().get_full_config()

    instance_service: InstanceService = service_factory.get_instance_service()

    instance_service.connect_to_rented_instance(
        instance_rented_slug=instance_uid,
        config=config,
        input_ssh_key_path=private_ssh_key_path
    )

    app_logger.info(f'Stop connecting to rented server instance')
    raise typer.Exit(0)


@self_hosted.command(name="ls", help=__("List self-hosted instances"))
def self_hosted_list(
        row: int = typer.Option(
            5,
            '--row',
            '-r',
            help=__("Set number of rows displayed per page"),
            is_eager=False,
        ),
        page: int = typer.Option(
            1,
            '--page',
            '-p',
            help=__("Set starting page for displaying output"),
            is_eager=False,
        ),
        statuses: List[str] = typer.Option(
            None,
            '--status',
            '-s',
            help=__("Filter by status, use --status all to list all self-hosted server instances"),
            is_eager=False,
        ),
):
    """
        Lists self-hosted instances
    """
    service_factory = validate_config_and_get_service_factory()
    config = service_factory.get_config_provider().get_full_config()

    instance_service: InstanceService = service_factory.get_instance_service()
    selfhosted_instance_status_map = service_factory.get_thestage_api_client().get_selfhosted_business_status_map(config.main.thestage_auth_token)

    if not statuses:
        statuses = ({key: selfhosted_instance_status_map[key] for key in [
            SelfhostedBusinessStatus.AWAITING_CONFIGURATION,
            SelfhostedBusinessStatus.RUNNING,
            SelfhostedBusinessStatus.TERMINATED,
        ]}).values()

    if "all" in statuses:
        statuses = selfhosted_instance_status_map.values()

    for input_status_item in statuses:
        if input_status_item not in selfhosted_instance_status_map.values():
            typer.echo(__("'%invalid_status%' is not one of %valid_statuses%", {
                'invalid_status': input_status_item,
                'valid_statuses': str(list(selfhosted_instance_status_map.values()))
            }))
            raise typer.Exit(1)

    typer.echo(__(
        "Listing self-hosted instances with the following statuses: %statuses%, to view all self-hosted instances, use --status all",
        placeholders={
            'statuses': ', '.join([status_item for status_item in statuses])
        }))

    backend_statuses: List[str] = [key for key, value in selfhosted_instance_status_map.items() if value in statuses]

    instance_service.print(
        func_get_data=instance_service.get_self_hosted_list,
        func_special_params={
            'statuses': backend_statuses,
        },
        mapper=SelfHostedMapper(),
        config=config,
        headers=list(map(lambda x: x.alias, SelfHostedInstanceEntity.model_fields.values())),
        row=row,
        page=page,
        show_index="never",
    )

    typer.echo(__("Self-hosted instances listing complete"))
    raise typer.Exit(0)


@self_hosted.command(name="connect", no_args_is_help=True, help=__("Connect to self-hosted instance"))
def self_hosted_connect(
        instance_slug: Optional[str] = typer.Argument(help=__("Self-hosted instance unique ID"),),
        username: Optional[str] = typer.Option(
            None,
            '--username',
            '-u',
            help=__("Username for server instance (required when connecting to self-hosted instance)"),
            is_eager=False,
        ),
        private_ssh_key_path: str = typer.Option(
            None,
            "--private-key-path",
            "-pk",
            help=__("Path to private key that will be accepted by remote server (optional)"),
            is_eager=False,
        ),
):
    """
        Connects to self-hosted instance
    """
    app_logger.info(f'Connect to self-hosted instance')

    if not instance_slug:
        typer.echo(__('Self-hosted instance unique ID is required'))
        raise typer.Exit(1)

    if private_ssh_key_path and not Path(private_ssh_key_path).is_file():
        typer.echo(f'No file found at provided path {private_ssh_key_path}')
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    config = service_factory.get_config_provider().get_full_config()

    instance_service: InstanceService = service_factory.get_instance_service()

    instance_service.connect_to_selfhosted_instance(
        selfhosted_instance_slug=instance_slug,
        username=username,
        config=config,
        input_ssh_key_path=private_ssh_key_path
    )

    app_logger.info(f'Stop connecting to self-hosted instance')
    raise typer.Exit(0)



from typing import Optional

from pydantic import Field, ConfigDict, BaseModel


class UserLogsQueryRequest(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    containerRunId: Optional[int] = Field(None, alias='containerRunId')
    containerId: Optional[int] = Field(None, alias='containerId')
    message: Optional[str] = Field(None, alias='message')
    instanceId: Optional[int] = Field(None, alias='instanceId')
    instanceType: Optional[str] = Field(None, alias='instanceType') # ENUM
    taskId: Optional[int] = Field(None, alias='taskId')
    inferenceSimulatorId: Optional[int] = Field(None, alias='inferenceSimulatorId')
    fromTimestamp: Optional[str] = Field(None, alias='fromTimestamp')
    toTimestamp: Optional[str] = Field(None, alias='toTimestamp')
    logType: Optional[str] = Field(None, alias='logType') # ENUM
    offset: Optional[int] = Field(None, alias='offset')
    limit: Optional[int] = Field(None, alias='limit')
    ascendingOrder: Optional[bool] = Field(None, alias='ascendingOrder')



import asyncio
from _signal import SIGINT
from asyncio import CancelledError, Task, StreamWriter
from datetime import datetime
from typing import Optional, Dict

import aioconsole
import typer
from httpx import ReadTimeout, ConnectError, ConnectTimeout
from requests.exceptions import ChunkedEncodingError

from thestage.helpers.logger.app_logger import app_logger
from thestage.services.clients.thestage_api.dtos.enums.container_status import DockerContainerStatus
from thestage.services.core_files.config_entity import ConfigEntity
from thestage.services.clients.thestage_api.dtos.enums.inference_simulator_status import InferenceSimulatorStatus
from thestage.services.clients.thestage_api.dtos.enums.task_status import TaskStatus
from thestage.services.clients.thestage_api.dtos.inference_controller.get_inference_simulator_response import \
    GetInferenceSimulatorResponse
from thestage.services.clients.thestage_api.dtos.task_controller.task_view_response import TaskViewResponse
from thestage.services.logging.byte_print_style import BytePrintStyle
from thestage.services.logging.dto.log_message import LogMessage
from thestage.services.logging.dto.log_type import LogType
from thestage.i18n.translation import __
from thestage.services.abstract_service import AbstractService
from thestage.services.clients.thestage_api.dtos.container_response import DockerContainerDto
from thestage.helpers.error_handler import error_handler
from thestage.services.clients.thestage_api.api_client import TheStageApiClient
from thestage.services.config_provider.config_provider import ConfigProvider
from rich import print

from thestage.services.logging.exception.log_polling_exception import LogPollingException
from thestage.services.logging.logging_constants import LOG_MESSAGE_CODE_TASK_FINISHED, \
    LOG_MESSAGE_CODE_INFERENCE_SIMULATOR_FAILED

is_logs_streaming = False


class LoggingService(AbstractService):
    __thestage_api_client: TheStageApiClient = None

    def __init__(
            self,
            thestage_api_client: TheStageApiClient,
            config_provider: ConfigProvider,
    ):
        super(LoggingService, self).__init__(
            config_provider=config_provider
        )
        self.__thestage_api_client = thestage_api_client


    @error_handler()
    def print_last_task_logs(self, config: ConfigEntity, task_id: int, logs_number: Optional[int]):
        logs = self.__thestage_api_client.query_user_logs(
            token=config.main.thestage_auth_token,
            task_id=task_id,
            limit=logs_number
        )
        for log_message in reversed(logs.queryResult):
            self.__print_log_line_object(log_message)


    @error_handler()
    def print_last_inference_simulator_logs(self, config: ConfigEntity, inference_simulator_id: int, logs_number: Optional[int]):
        logs = self.__thestage_api_client.query_user_logs(
            token=config.main.thestage_auth_token,
            inference_simulator_id=inference_simulator_id,
            limit=logs_number
        )
        for log_message in reversed(logs.queryResult):
            self.__print_log_line_object(log_message)


    @error_handler()
    def print_last_container_logs(self, config: ConfigEntity, container_uid: str, logs_number: Optional[int]):
        container: Optional[DockerContainerDto] = self.__thestage_api_client.get_container(
            token=config.main.thestage_auth_token,
            container_slug=container_uid,
        )

        if not container:
            typer.echo("Container was not found")
            raise typer.Exit(1)

        logs = self.__thestage_api_client.query_user_logs(
            token=config.main.thestage_auth_token,
            container_id=container.id,
            limit=logs_number
        )
        for log_message in reversed(logs.queryResult):
            self.__print_log_line_object(log_message)


    @error_handler()
    def stream_task_logs_with_controls(self, config: ConfigEntity, task_id: int):
        asyncio.run(
            self.__stream_task_logs_with_controls_async(
                config=config,
                task_id=task_id
            )
        )


    @error_handler()
    async def __stream_task_logs_with_controls_async(self, config: ConfigEntity, task_id: int):
        task_view_response: Optional[TaskViewResponse] = self.__thestage_api_client.get_task(
            token=config.main.thestage_auth_token,
            task_id=task_id,
        )

        task_status_map: Dict[str, str] = self.__thestage_api_client.get_task_localized_status_map(
            token=config.main.thestage_auth_token,
        )

        task = task_view_response.task

        if task:
            if task.frontend_status.status_key not in [TaskStatus.RUNNING, TaskStatus.SCHEDULED]:
                typer.echo(__("Task must be in status '%required_status%' to stream real-time logs. Task %task_id% status: '%status%'.", {
                    'task_id': str(task.id),
                    'status': task.frontend_status.status_translation,
                    'required_status': task_status_map.get(TaskStatus.RUNNING) or TaskStatus.RUNNING
                }))
                raise typer.Exit(1)
        else:
            typer.echo(__("Task with ID %task_id% was not found", {'task_id': task.id}))
            raise typer.Exit(1)

        typer.echo(__(
            f"Log stream for task %task_id% started",
            {
                'task_id': str(task.id),
            }
        ))

        typer.echo(__("CTRL+C to cancel the task. CTRL+D to disconnect from log stream."))

        print_logs_task = asyncio.create_task(self.print_realtime_logs(token=config.main.thestage_auth_token, task_id=task.id))
        input_task = asyncio.create_task(self.read_log_stream_input())

        def sigint_handler():
            input_task.cancel()

        loop = asyncio.get_event_loop()
        for signal_item in [SIGINT]:  # SIGINT == CTRL+C
            loop.add_signal_handler(signal_item, sigint_handler)

        done, pending = await asyncio.wait([print_logs_task, input_task], return_when=asyncio.FIRST_COMPLETED)

        if input_task in done:
            print_logs_task.cancel()
            if not input_task.result():  # result is only expected if ctrl+D triggered EOFError
                typer.echo(f"\rTask {task_id} will be canceled")
                self.__thestage_api_client.cancel_task(
                    token=config.main.thestage_auth_token,
                    task_id=task.id,
                )

    @error_handler()
    def stream_inference_simulator_logs_with_controls(self, config: ConfigEntity, slug: str):
        asyncio.run(
            self.__stream_inference_simulator_logs_with_controls_async(
                config=config,
                slug=slug
            )
        )

    @error_handler()
    async def __stream_inference_simulator_logs_with_controls_async(self, config: ConfigEntity, slug: str):
        get_inference_simulator_response: Optional[GetInferenceSimulatorResponse] = self.__thestage_api_client.get_inference_simulator(
            token=config.main.thestage_auth_token,
            slug=slug,
        )

        inference_simulator_status_map: Dict[str, str] = self.__thestage_api_client.get_inference_simulator_business_status_map(
            token=config.main.thestage_auth_token,
        )

        inference_simulator = get_inference_simulator_response.inferenceSimulator

        if inference_simulator:
            if inference_simulator.status not in ['SCHEDULED', 'CREATING', 'RUNNING']:
                typer.echo(
                    __("Inference simulator must be in status '%required_status%' to stream real-time logs. Inference simulator status: '%status%'.",
                       {
                           'status': inference_simulator.status,
                           'required_status': inference_simulator_status_map.get(InferenceSimulatorStatus.RUNNING) or InferenceSimulatorStatus.RUNNING
                       }))
                raise typer.Exit(1)
        else:
            typer.echo(__("Inference simulator with unique ID %slug% was not found", {'slug': inference_simulator.slug}))
            raise typer.Exit(1)

        typer.echo(__(
            f"Log stream for inference simulator %slug% started",
            {
                'slug': str(inference_simulator.slug),
            }
        ))

        typer.echo(__("CTRL+D to disconnect from log stream."))

        print_task_or_inference_simulator_logs = asyncio.create_task(
            self.print_realtime_logs(token=config.main.thestage_auth_token, inference_simulator_id=inference_simulator.id)
        )
        input_task = asyncio.create_task(self.read_log_stream_input())

        done, pending = await asyncio.wait([print_task_or_inference_simulator_logs, input_task],
                                           return_when=asyncio.FIRST_COMPLETED)

        if input_task in done:
            print_task_or_inference_simulator_logs.cancel()
            typer.echo(__(f"Disconnected from log stream. You can try to reconnect with 'thestage project inference-simulator logs {slug}'."))


    @error_handler()
    def stream_container_logs_with_controls(self, config: ConfigEntity, container_uid: str):
        asyncio.run(
            self.__stream_container_logs_with_controls_async(
                config=config,
                container_uid=container_uid
            )
        )


    @error_handler()
    async def __stream_container_logs_with_controls_async(self, config: ConfigEntity, container_uid: str):
        container: Optional[DockerContainerDto] = self.__thestage_api_client.get_container(
            token=config.main.thestage_auth_token,
            container_slug=container_uid,
        )

        if container:
            if container.frontend_status.status_key not in [DockerContainerStatus.RUNNING]:
                typer.echo(f"Container status: '{container.frontend_status.status_translation}'")
        else:
            typer.echo("Container was not found")
            raise typer.Exit(1)

        typer.echo(f"Log stream for Docker container started")
        typer.echo("CTRL+D to disconnect from log stream.")

        print_logs_task = asyncio.create_task(self.print_realtime_logs(token=config.main.thestage_auth_token, docker_container_id=container.id))
        input_task = asyncio.create_task(self.read_log_stream_input())

        def sigint_handler():
            input_task.cancel()

        loop = asyncio.get_event_loop()
        for signal_item in [SIGINT]:  # SIGINT == CTRL+C
            loop.add_signal_handler(signal_item, sigint_handler)

        done, pending = await asyncio.wait([print_logs_task, input_task], return_when=asyncio.FIRST_COMPLETED)

        if input_task in done:
            print_logs_task.cancel()


    async def read_log_stream_input(self):
        try:
            while True:
                input1 = await aioconsole.ainput()
        except EOFError:
            typer.echo(__("\rExited from log stream"))
            return True
        except CancelledError:  # Always appears if async task is canceled and leaves huge traces
            pass


    async def print_realtime_logs(
            self,
            token: str,
            task_id: Optional[int] = None,
            inference_simulator_id: Optional[int] = None,
            docker_container_id: Optional[int] = None,
    ):
        polling_interval_seconds: float = 4  # also adjust polling api method timeout if changed
        between_logs_sleeping_coef: float = 1  # we emulate delay between logs, but if for any reason code runs for too long - delays will be controlled with this coef
        last_iteration_log_timestamp: Optional[str] = None  # pointer to next iteration polling start (obtained from each response)
        last_log_id: Optional[str] = None  # pointer to next iteration polling start - to exclude the log id from result (obtained from each response)
        consecutive_error_count: int = 0  # connectivity errors count - stream will disconnect if too many errors in a row
        iteration_started_at: datetime  # used to control iteration duration - polling should be done at around exact rate
        errors_started_at: Optional[datetime] = None  # time since errors started to stream disconnect

        is_no_more_logs = False
        while not is_no_more_logs:
            log_wait_remaining_limit: float = polling_interval_seconds  # hard limit just in case
            iteration_started_at = datetime.utcnow()
            last_printed_log_timestamp: Optional[datetime] = None
            reader, writer = await aioconsole.get_standard_streams()

            # this shows (somewhat accurate) time difference between logs here and in real time. should not grow.
            # if last_iteration_log_timestamp:
            #     last_log_timestamp_parsed = datetime.strptime(last_iteration_log_timestamp, '%Y-%m-%dT%H:%M:%S.%f')
            #     stream_to_logs_diff = datetime.utcnow() - last_log_timestamp_parsed
            #     print_nonblocking(f'TDIFF {stream_to_logs_diff.total_seconds()}', writer)
            try:
                logs_response = await self.__thestage_api_client.poll_logs_httpx(
                    token=token,
                    task_id=task_id,
                    inference_simulator_id=inference_simulator_id,
                    docker_container_id=docker_container_id,
                    last_log_timestamp=last_iteration_log_timestamp,
                    last_log_id=last_log_id
                )

                if not logs_response.is_success:
                    app_logger.info(f'Polling logs error: {logs_response.message}')
                    raise LogPollingException('')


                if consecutive_error_count > 0:
                    consecutive_error_count = 0
                    errors_started_at = None
                    log_wait_remaining_limit = 0   # no log delays after reconnect

                last_iteration_log_timestamp = logs_response.lastLogTimestamp
                last_log_id = logs_response.lastLogId

                for log_item in logs_response.logs:
                    current_log_timestamp = datetime.strptime(log_item.timestamp[:26], '%Y-%m-%dT%H:%M:%S.%f')  # python does not like nanoseconds
                    if last_printed_log_timestamp is not None and log_wait_remaining_limit > 0:
                        logs_sleeptime = (current_log_timestamp - last_printed_log_timestamp).total_seconds() * between_logs_sleeping_coef
                        await asyncio.sleep(logs_sleeptime)
                        log_wait_remaining_limit -= logs_sleeptime
                    self.__print_log_line_object_nonblocking(log_item, writer)
                    last_printed_log_timestamp = current_log_timestamp
                    if log_item.messageCode == LOG_MESSAGE_CODE_TASK_FINISHED or log_item.messageCode == LOG_MESSAGE_CODE_INFERENCE_SIMULATOR_FAILED:
                        is_no_more_logs = True

                if is_no_more_logs:
                    break
            except (ReadTimeout, ConnectError, ConnectTimeout, LogPollingException) as e:
                consecutive_error_count += 1
                if consecutive_error_count == 1:
                    if isinstance(e, LogPollingException):
                        print_nonblocking("Some problems raised while getting logs...", writer, BytePrintStyle.ORANGE)
                    else:
                        print_nonblocking("Network issues, attempting to re-establish connection...", writer, BytePrintStyle.ORANGE)
                if not errors_started_at:
                    errors_started_at = datetime.utcnow()

            if consecutive_error_count > 7:
                seconds_with_error = (datetime.utcnow() - errors_started_at).total_seconds()
                if inference_simulator_id:
                    print_nonblocking(f"Log stream: disconnected from server (connectivity issues for {seconds_with_error} seconds). Try 'thestage inference-simulator logs <inference-simulator-UID>' to reconnect.", writer)
                elif task_id:
                    print_nonblocking(f"Log stream: disconnected from server (connectivity issues for {seconds_with_error} seconds). Try 'thestage project task logs {task_id}' to reconnect.", writer)
                elif docker_container_id:
                    print_nonblocking(f"Log stream: disconnected from server (connectivity issues for {seconds_with_error} seconds). Try 'thestage container logs <docker-container-UID>' to reconnect.", writer)
                else:
                    print_nonblocking(f"Log stream: disconnected from server (connectivity issues for {seconds_with_error} seconds)", writer)
                break

            # depending on iteration duration - sleep for the remaining time and adjust log sleep coefficient if needed
            iteration_duration = (datetime.utcnow() - iteration_started_at).total_seconds()
            if iteration_duration > polling_interval_seconds:
                between_logs_sleeping_coef *= 0.85
            else:
                await asyncio.sleep(polling_interval_seconds - iteration_duration)
                if between_logs_sleeping_coef < 1:
                    between_logs_sleeping_coef = min(1.0, between_logs_sleeping_coef * 1.15)


    def __print_log_line(self, log_message_raw_json: str):
        log_message = LogMessage.model_validate_json(log_message_raw_json)
        if not log_message.logType and log_message.message == 'ping':
            return
        self.__print_log_line_object(log_message)


    @staticmethod
    def __print_log_line_object(log_message: LogMessage):
        line_color: str = "grey78"

        if not log_message.logType and log_message.message == 'ping':
            return

        if log_message.logType == LogType.STDERR.value:
            line_color = "red"
        if log_message.message:
            print(f'[{line_color}][not bold]{log_message.message}[/not bold][/{line_color}]')

    @staticmethod
    def __print_log_line_object_nonblocking(log_message: LogMessage, writer: StreamWriter):
        if log_message.message:
            line_color: str = BytePrintStyle.RESET
            if log_message.logType == LogType.STDERR.value:
                line_color = BytePrintStyle.RED

            print_nonblocking(f'{log_message.message}', writer, line_color)


def print_nonblocking(line: str, writer: StreamWriter, color_code: str = BytePrintStyle.RESET):
    writer.write(str.encode(f'{color_code}{line}{BytePrintStyle.RESET}\r\n'))



from typing import Optional

from pydantic import Field

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage.services.task.dto.task_dto import TaskDto


class TaskViewResponse(TheStageBaseResponse):
    task: Optional[TaskDto] = Field(None, alias='task')
    task_view_url: Optional[str] = Field(None, alias='websiteTaskViewUrl')
    task_output_url: Optional[str] = Field(None, alias='websiteTaskOutputUrl')



from typing import Optional, List

from pydantic import Field, BaseModel, ConfigDict

from thestage.services.clients.thestage_api.dtos.enums.container_pending_action import DockerContainerAction

class DockerContainerActionRequestDto(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    container_id: Optional[int] = Field(None, alias='dockerContainerId')
    action: DockerContainerAction = Field(None, alias='action')



from typing import List, Optional

from pydantic import Field, ConfigDict, BaseModel

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage.services.clients.thestage_api.dtos.container_response import DockerContainerDto
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList
from thestage.services.clients.thestage_api.dtos.pagination_data import PaginationData

class DockerContainerListResponse(TheStageBaseResponse):
    model_config = ConfigDict(use_enum_values=True)

    paginatedList: PaginatedEntityList[DockerContainerDto] = Field(None, alias='paginatedList')



import time
from pathlib import Path
from typing import Optional, List

import re

import typer
from typing_extensions import Annotated

from thestage.controllers.utils_controller import validate_config_and_get_service_factory, get_current_directory
from thestage.entities.project_inference_simulator import ProjectInferenceSimulatorEntity
from thestage.entities.project_inference_simulator_model import ProjectInferenceSimulatorModelEntity
from thestage.entities.project_task import ProjectTaskEntity
from thestage.helpers.logger.app_logger import app_logger
from thestage.i18n.translation import __
from thestage.services.clients.thestage_api.dtos.enums.inference_model_status import InferenceModelStatus
from thestage.services.clients.thestage_api.dtos.enums.inference_simulator_status import InferenceSimulatorStatus
from thestage.services.clients.thestage_api.dtos.inference_controller.get_inference_simulator_response import \
    GetInferenceSimulatorResponse
from thestage.services.logging.logging_service import LoggingService
from thestage.services.project.dto.project_config import ProjectConfig
from thestage.services.project.mapper.project_inference_simulator_mapper import ProjectInferenceSimulatorMapper
from thestage.services.project.mapper.project_inference_simulator_model_mapper import \
    ProjectInferenceSimulatorModelMapper
from thestage.services.project.mapper.project_task_mapper import ProjectTaskMapper
from thestage.services.project.project_service import ProjectService
from thestage.services.task.dto.task_dto import TaskDto

app = typer.Typer(no_args_is_help=True, help=__("Manage projects"))
inference_simulators_app = typer.Typer(no_args_is_help=True, help="Manage project inference simulators")
app.add_typer(inference_simulators_app, name="inference-simulator")
inference_simulator_model_app = typer.Typer(no_args_is_help=True, help="Manage project inference simulator models")
app.add_typer(inference_simulator_model_app, name="model")
task_app = typer.Typer(no_args_is_help=True, help=__("Manage project tasks"))
config_app = typer.Typer(no_args_is_help=True, help=__("Manage project config"))
app.add_typer(task_app, name="task")
app.add_typer(config_app, name="config")


@app.command(name='clone', no_args_is_help=True, help=__("Clone project repository to empty directory"))
def clone(
        project_uid: str = typer.Argument(
            help=__("Project unique ID"),
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to the working directory: current directory used by default"),
            is_eager=False,
        ),
):
    """
        Clones project to current working directory
    """
    app_logger.info(f'Start project clone from {get_current_directory()}')
    if not working_directory:
        working_directory = get_current_directory().joinpath(project_uid)

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    config = service_factory.get_config_provider().get_full_config()

    project_service = service_factory.get_project_service()

    project_service.clone_project(
        config=config,
        project_slug=project_uid,
    )

    typer.echo(__("Project successfully cloned to %path%", {"path": config.runtime.working_directory}))
    raise typer.Exit(0)


@app.command(name='init', no_args_is_help=True, help=__("Initialize project repository with existing files"))
def init(
        project_uid: Optional[str] = typer.Argument(
            help=__("Project unique ID"),
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
):
    """
        Initializes project in current working directory
    """
    app_logger.info(f'Start project init from {get_current_directory()}')

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    config = service_factory.get_config_provider().get_full_config()

    project_service = service_factory.get_project_service()
    project_config = service_factory.get_config_provider().read_project_config()

    if project_config:
        typer.echo(__("Directory is initialized and already contains working project"))
        raise typer.Exit(1)

    project_service.init_project(
        project_slug=project_uid,
        config=config,
    )

    typer.echo(__("Project successfully initialized at %path%", {"path": config.runtime.working_directory}))
    raise typer.Exit(0)


@app.command(name='run', no_args_is_help=True, help=__("Run a task within the project. By default, it uses the latest commit from the main branch and streams real-time task logs."))
def run(
        command: Annotated[List[str], typer.Argument(
            help=__("Command to run (required)"),
        )],
        commit_hash: Optional[str] = typer.Option(
            None,
            '--commit-hash',
            '-hash',
            help=__("Commit hash to use. By default, the current HEAD commit is used."),
            is_eager=False,
        ),
        docker_container_slug: Optional[str] = typer.Option(
            None,
            '--container-uid',
            '-cid',
            help=__("Docker container unique ID"),
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            show_default=False,
            is_eager=False,
        ),
        enable_log_stream: Optional[bool] = typer.Option(
            True,
            " /--no-logs",
            " /-nl",
            help=__("Disable real-time log streaming"),
            is_eager=False,
        ),
        task_title: Optional[str] = typer.Option(
            None,
            "--title",
            "-t",
            help=__("Provide a custom task title. Git commit message is used by default."),
            is_eager=False,
        ),
):
    """
        Runs a task within a project
    """
    app_logger.info(f'Start project run from {get_current_directory()}')

    if not command:
        typer.echo(__('Command is required'))
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    config = service_factory.get_config_provider().get_full_config()

    project_service = service_factory.get_project_service()

    task: Optional[TaskDto] = project_service.project_run_task(
        config=config,
        run_command=" ".join(command),
        commit_hash=commit_hash,
        docker_container_slug=docker_container_slug,
        task_title=task_title,
    )

    if enable_log_stream:
        logging_service: LoggingService = service_factory.get_logging_service()

        logging_service.stream_task_logs_with_controls(
            config=config,
            task_id=task.id
        )

    raise typer.Exit(0)


@task_app.command(name='cancel', no_args_is_help=True, help=__("Cancel a task by ID"))
def cancel_task(
        task_id: Annotated[int, typer.Argument(
            help=__("Task ID (required)"),
        )],
):
    """
        Cancels a task
    """
    app_logger.info(f'Start cancel task from {get_current_directory()}')

    if not task_id:
        typer.echo(__('Task ID is required'))
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    config = service_factory.get_config_provider().get_full_config()

    project_service = service_factory.get_project_service()

    project_service.cancel_task(
        config=config,
        task_id=task_id
    )

    raise typer.Exit(0)


@task_app.command("ls", help=__("List tasks"))
def list_runs(
        project_uid: Annotated[str, typer.Argument(help=__("Project unique ID. By default, project info is taken from the current directory"), metavar="OPTIONAL")] = None,
        row: int = typer.Option(
            5,
            '--row',
            '-r',
            help=__("Set number of rows displayed per page"),
            is_eager=False,
        ),
        page: int = typer.Option(
            1,
            '--page',
            '-p',
            help=__("Set starting page for displaying output"),
            is_eager=False,
        ),
):
    """
        Lists tasks for a project
    """
    app_logger.info(f'Start project list-runs from {get_current_directory()}')

    service_factory = validate_config_and_get_service_factory()
    config = service_factory.get_config_provider().get_full_config()

    project_service: ProjectService = service_factory.get_project_service()

    if not project_uid:
        project_config: ProjectConfig = service_factory.get_config_provider().read_project_config()
        if not project_config:
            typer.echo(__("Provide the project unique ID or run this command from within an initialized project directory"))
            raise typer.Exit(1)
        project_uid = project_config.slug

    project_service.print(
        func_get_data=project_service.get_project_task_list,
        func_special_params={
            'project_slug': project_uid,
        },
        mapper=ProjectTaskMapper(),
        config=config,
        headers=list(map(lambda x: x.alias, ProjectTaskEntity.model_fields.values())),
        row=row,
        page=page,
        max_col_width=[100, 100, 100, 100, 100, 100, 100, 100],
        show_index="never",
    )

    typer.echo(__("Tasks listing complete"))
    raise typer.Exit(0)


@task_app.command(name="logs", no_args_is_help=True, help=__("Stream real-time task logs or view last logs for a task"))
def task_logs(
        task_id: Optional[int] = typer.Argument(help=__("Task ID"),),
        logs_number: Optional[int] = typer.Option(
            None,
            '--number',
            '-n',
            help=__("Display a number of latest log entries. No real-time stream if provided."),
            is_eager=False,
        ),
):
    """
        Streams real-time task logs
    """
    app_logger.info(f'View task logs')

    if not task_id:
        typer.echo(__('Task ID is required'))
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    config = service_factory.get_config_provider().get_full_config()

    logging_service: LoggingService = service_factory.get_logging_service()

    if logs_number is None:
        logging_service.stream_task_logs_with_controls(
            config=config,
            task_id=task_id
        )
    else:
        logging_service.print_last_task_logs(config=config, task_id=task_id, logs_number=logs_number)

    app_logger.info(f'Task logs - end')
    raise typer.Exit(0)


@app.command(name='checkout', no_args_is_help=True, help=__("Checkout project repository to a specific reference"))
def checkout_project(
        reference: Optional[str] = typer.Argument(
            help=__("Task ID or branch name to checkout. '/' will try to identify the main branch."),
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
):
    """
        Checkout project in current working directory
    """
    app_logger.info(f'Start project checkout from {get_current_directory()}')

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    config = service_factory.get_config_provider().get_full_config()

    project_service = service_factory.get_project_service()

    task_id: Optional[int] = None
    branch_name: Optional[str] = None

    if reference == "/":
        pass
    elif reference.isdigit():
        task_id = int(reference)
    else:
        branch_name = reference

    project_service.checkout_project(
        config=config,
        task_id=task_id,
        branch_name=branch_name
    )

    raise typer.Exit(0)


@app.command(name='pull', help=__("Pulls the changes from the remote project repository. Equivalent to 'git pull'."))
def pull_project(
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
):
    """
        Pull project in current working directory
    """
    app_logger.info(f'Start project pull from {get_current_directory()}')

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    config = service_factory.get_config_provider().get_full_config()

    project_service = service_factory.get_project_service()

    project_service.pull_project(
        config=config,
    )

    raise typer.Exit(0)


@app.command(name='reset', help=__("Resets the current project branch to remote counterpart. All working tree changes will be lost. Equivalent to 'git fetch && git reset --hard origin/{ref}'."))
def reset_project(
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
):
    """
        Pull project in current working directory
    """
    app_logger.info(f'Start project pull from {get_current_directory()}')

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    config = service_factory.get_config_provider().get_full_config()

    project_service = service_factory.get_project_service()

    project_service.reset_project(
        config=config
    )

    raise typer.Exit(0)


@config_app.command(name='set-default-container', no_args_is_help=True, help=__("Set default docker container for a project installation"))
def set_default_container(
        container_uid: Annotated[Optional[str], typer.Argument(
            help=__("Unique ID of the container to use by default for running tasks"),
        )] = None,
        unset_default_container: Optional[bool] = typer.Option(
            False,
            "--unset",
            "-u",
            help=__("Unsets the default docker container"),
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
):
    """
        Initializes project in current working directory
    """
    app_logger.info(f'Start project init from {get_current_directory()}')

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    config = service_factory.get_config_provider().get_full_config()

    if unset_default_container and container_uid:
        typer.echo("Container unique ID is provided along with unset flag. Please pick one.")
        raise typer.Exit(1)

    if not unset_default_container and not container_uid:
        typer.echo("Provide container unique ID or use '--unset' flag")
        raise typer.Exit(1)

    project_service = service_factory.get_project_service()

    project_service.set_default_container(
        config=config,
        container_uid=container_uid,
    )

    raise typer.Exit(0)


@config_app.command(name='get', no_args_is_help=False, help=__("View config for a local project installation"))
def get_project_config(
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
):
    """
        Initializes project in current working directory
    """
    app_logger.info(f'Start project init from {get_current_directory()}')

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    config = service_factory.get_config_provider().get_full_config()

    project_service = service_factory.get_project_service()

    project_service.print_project_config(
        config=config,
    )

    raise typer.Exit(0)


@inference_simulators_app.command(name='run', no_args_is_help=True, help="Run an inference simulator within the project")
def run_inference_simulator(
        unique_id: Optional[str] = typer.Argument(help=__("Inference simulator unique ID"), ),
        rented_instance_unique_id: Optional[str] = typer.Option(
            None,
            '--rented-instance-unique-id',
            '-ruid',
            help=__("The rented instance unique ID on which the inference simulator will run"),
            is_eager=False,
        ),
        self_hosted_instance_unique_id: Optional[str] = typer.Option(
            None,
            '--self-hosted-instance-unique-id',
            '-suid',
            help=__("The self-hosted instance unique ID on which the inference simulator will run"),
            is_eager=False,
        ),
        commit_hash: Optional[str] = typer.Option(
            None,
            '--commit-hash',
            '-hash',
            help=__("Commit hash to use. By default, the current HEAD commit is used."),
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory. By default, the current directory is used"),
            show_default=False,
            is_eager=False,
        ),
        enable_log_stream: Optional[bool] = typer.Option(
            True,
            " /--no-logs",
            " /-nl",
            help=__("Disable real-time log streaming"),
            is_eager=False,
        ),
        is_skip_installation: Optional[bool] = typer.Option(
            False,
            "--skip-installation",
            "-si",
            help=__("Skip installing dependencies from requirements.txt and install.sh"),
            is_eager=False,
        ),
):
    """
        Run an inference simulator within a project
    """

    if unique_id and not re.match(r"^[a-zA-Z0-9-]+$", unique_id):
        raise typer.BadParameter(__("Invalid UID format. The UID can only contain letters, numbers, and hyphens."))

    app_logger.info(f'Running an inference simulator from {get_current_directory()}')

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    config = service_factory.get_config_provider().get_full_config()

    working_dir_path = Path(working_directory) if working_directory else Path(config.runtime.working_directory)
    inference_files = list(working_dir_path.rglob("inference.py"))
    if not inference_files:
        typer.echo("No inference.py file found in the project directory.")
        raise typer.Exit(1)
    elif len(inference_files) == 1:
        selected_inference = inference_files[0]
    else:
        choices = [str(path.relative_to(working_dir_path)) for path in inference_files]
        typer.echo("Multiple inference.py files found:")
        for idx, choice in enumerate(choices, start=1):
            typer.echo(f"{idx}) {choice}")
        choice_str = typer.prompt("Choose which inference.py to use")
        try:
            choice_index = int(choice_str)
        except ValueError:
            raise typer.BadParameter("Invalid input. Please enter a number.")
        if not (1 <= choice_index <= len(choices)):
            raise typer.BadParameter("Choice out of range.")
        selected_inference = inference_files[choice_index - 1]

    relative_inference = selected_inference.relative_to(working_dir_path)
    parent_dir = relative_inference.parent
    if parent_dir == Path("."):
        inference_dir = "/"
    else:
        inference_dir = f"{parent_dir.as_posix()}/"
    typer.echo(f"Selected inference file relative path: {inference_dir}")

    project_service = service_factory.get_project_service()

    project_service.project_run_inference_simulator(
        config=config,
        commit_hash=commit_hash,
        slug=unique_id,
        rented_instance_unique_id=rented_instance_unique_id,
        self_hosted_instance_unique_id=self_hosted_instance_unique_id,
        inference_dir=inference_dir,
        is_skip_installation=is_skip_installation,
    )

    if enable_log_stream:
        logging_service: LoggingService = service_factory.get_logging_service()

        logging_service.stream_inference_simulator_logs_with_controls(
            config=config,
            slug=unique_id
        )
    raise typer.Exit(0)


@inference_simulators_app.command(name='save-metadata', no_args_is_help=True, help="Get and save inference simulator metadata")
def get_and_save_inference_simulator_metadata(
        unique_id: Optional[str] = typer.Argument(help=__("Inference simulator unique ID"),),
        file_path: Optional[str] = typer.Option(
            None,
            "--file-path",
            "-fp",
            help=__("Full path to a new file. By default metadata is saved to the current directory as metadata.json"),
            show_default=False,
            is_eager=False,
        ),
):
    """
        Get and save inference simulator metadata
    """

    service_factory = validate_config_and_get_service_factory()
    config = service_factory.get_config_provider().get_full_config()

    project_service = service_factory.get_project_service()

    project_service.project_get_and_save_inference_simulator_metadata(
        config=config,
        file_path=file_path,
        slug=unique_id,
    )

    raise typer.Exit(0)


@inference_simulators_app.command(name='push', no_args_is_help=True, help="Push an inference simulator within the project to model registry")
def push_inference_simulator(
        unique_id: Optional[str] = typer.Argument(help=__("Inference simulator unique ID"),),
):
    """
        Push an inference simulator within a project to AWS
    """
    app_logger.info(f'Pushing an inference simulator')

    service_factory = validate_config_and_get_service_factory()
    config = service_factory.get_config_provider().get_full_config()

    project_service = service_factory.get_project_service()

    project_service.project_push_inference_simulator(
        config=config,
        slug=unique_id,
    )

    raise typer.Exit(0)


@inference_simulators_app.command("ls", help=__("List inference simulators"))
def list_inference_simulators(
        project_uid: Annotated[str, typer.Argument(help=__("Project unique ID. By default, project info is taken from the current directory"), metavar="OPTIONAL")] = None,
        row: int = typer.Option(
            5,
            '--row',
            '-r',
            help=__("Set number of rows displayed per page"),
            is_eager=False,
        ),
        page: int = typer.Option(
            1,
            '--page',
            '-p',
            help=__("Set starting page for displaying output"),
            is_eager=False,
        ),
        statuses: List[str] = typer.Option(
            None,
            '--status',
            '-s',
            help=__("Filter by status, use --status all to list all inference simulators"),
            is_eager=False,
        ),
):
    """
        Lists inference simulators for a project
    """
    app_logger.info(f'Start project list inference simulators from {get_current_directory()}')

    service_factory = validate_config_and_get_service_factory()
    config = service_factory.get_config_provider().get_full_config()

    project_service: ProjectService = service_factory.get_project_service()
    if not project_uid:
        project_config: ProjectConfig = service_factory.get_config_provider().read_project_config()
        if not project_config:
            typer.echo(__("Provide the project unique ID or run this command from within an initialized project directory"))
            raise typer.Exit(1)
        project_uid = project_config.slug

    inference_simulator_status_map = service_factory.get_thestage_api_client().get_inference_simulator_business_status_map(
        config.main.thestage_auth_token)

    if not statuses:
        statuses = ({key: inference_simulator_status_map[key] for key in [
            InferenceSimulatorStatus.SCHEDULED,
            InferenceSimulatorStatus.CREATING,
            InferenceSimulatorStatus.RUNNING,
        ]}).values()

    if "all" in statuses:
        statuses = inference_simulator_status_map.values()

    for input_status_item in statuses:
        if input_status_item not in inference_simulator_status_map.values():
            typer.echo(__("'%invalid_status%' is not one of %valid_statuses%", {
                'invalid_status': input_status_item,
                'valid_statuses': str(list(inference_simulator_status_map.values()))
            }))
            raise typer.Exit(1)

    typer.echo(__(
        "Listing inference simulators with the following statuses: %statuses%, to view all inference simulators, use --status all",
        placeholders={
            'statuses': ', '.join([status_item for status_item in statuses])
        }))

    backend_statuses: List[str] = [key for key, value in inference_simulator_status_map.items() if value in statuses]

    project_service.print(
        func_get_data=project_service.get_project_inference_simulator_list,
        func_special_params={
            'project_slug': project_uid,
            'statuses': backend_statuses,
        },
        mapper=ProjectInferenceSimulatorMapper(),
        config=config,
        headers=list(map(lambda x: x.alias, ProjectInferenceSimulatorEntity.model_fields.values())),
        row=row,
        page=page,
        max_col_width=[100, 100, 100, 100, 100, 100, 100, 100],
        show_index="never",
    )

    typer.echo(__("Inference simulators listing complete"))
    raise typer.Exit(0)


@inference_simulator_model_app.command("ls", help=__("List inference simulator models"))
def list_inference_simulator_models(
        project_uid: Annotated[str, typer.Argument(help=__("Project unique ID. By default, project info is taken from the current directory"), metavar="OPTIONAL")] = None,
        row: int = typer.Option(
            5,
            '--row',
            '-r',
            help=__("Set number of rows displayed per page"),
            is_eager=False,
        ),
        page: int = typer.Option(
            1,
            '--page',
            '-p',
            help=__("Set starting page for displaying output"),
            is_eager=False,
        ),
        statuses: List[str] = typer.Option(
            None,
            '--status',
            '-s',
            help=__("Filter by status, use --status all to list all inference simulator models"),
            is_eager=False,
        ),
):
    """
        Lists inference simulator models for a project
    """
    app_logger.info(f'Start project list inference simulator models from {get_current_directory()}')

    service_factory = validate_config_and_get_service_factory()
    config = service_factory.get_config_provider().get_full_config()

    project_service: ProjectService = service_factory.get_project_service()
    if not project_uid:
        project_config: ProjectConfig = service_factory.get_config_provider().read_project_config()
        if not project_config:
            typer.echo(__("Provide the project unique ID or run this command from within an initialized project directory"))
            raise typer.Exit(1)
        project_uid = project_config.slug

    inference_simulator_model_status_map = service_factory.get_thestage_api_client().get_inference_simulator_model_business_status_map(
        config.main.thestage_auth_token)

    if not statuses:
        statuses = ({key: inference_simulator_model_status_map[key] for key in [
            InferenceModelStatus.SCHEDULED,
            InferenceModelStatus.PROCESSING,
            InferenceModelStatus.PUSH_SUCCEED,
        ]}).values()

    if "all" in statuses:
        statuses = inference_simulator_model_status_map.values()

    for input_status_item in statuses:
        if input_status_item not in inference_simulator_model_status_map.values():
            typer.echo(__("'%invalid_status%' is not one of %valid_statuses%", {
                'invalid_status': input_status_item,
                'valid_statuses': str(list(inference_simulator_model_status_map.values()))
            }))
            raise typer.Exit(1)

    typer.echo(__(
        "Listing inference simulator models with the following statuses: %statuses%, to view all inference simulator models, use --status all",
        placeholders={
            'statuses': ', '.join([status_item for status_item in statuses])
        }))

    backend_statuses: List[str] = [key for key, value in inference_simulator_model_status_map.items() if value in statuses]

    project_service.print(
        func_get_data=project_service.get_project_inference_simulator_model_list,
        func_special_params={
            'project_slug': project_uid,
            'statuses': backend_statuses,
        },
        mapper=ProjectInferenceSimulatorModelMapper(),
        config=config,
        headers=list(map(lambda x: x.alias, ProjectInferenceSimulatorModelEntity.model_fields.values())),
        row=row,
        page=page,
        max_col_width=[100, 100, 100, 100, 100, 100, 100, 100],
        show_index="never",
    )

    typer.echo(__("Inference simulator models listing complete"))
    raise typer.Exit(0)


@inference_simulators_app.command(name="logs", no_args_is_help=True, help=__("Stream real-time task logs or view last logs for an inference simulator"))
def inference_simulator_logs(
        unique_id: Optional[str] = typer.Argument(help=__("Inference simulator unique ID"),),
        logs_number: Optional[int] = typer.Option(
            None,
            '--number',
            '-n',
            help=__("Display a number of latest log entries. No real-time stream if provided."),
            is_eager=False,
        ),
):
    """
        Streams real-time inference simulator logs
    """
    app_logger.info(f'View inference simulator logs')

    if not unique_id:
        typer.echo(__('Inference simulator UID is required'))
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    config = service_factory.get_config_provider().get_full_config()

    logging_service: LoggingService = service_factory.get_logging_service()

    if logs_number is None:
        logging_service.stream_inference_simulator_logs_with_controls(
            config=config,
            slug=unique_id
        )
    else:
        get_inference_simulator_response: Optional[GetInferenceSimulatorResponse] = service_factory.get_thestage_api_client().get_inference_simulator(
            token=config.main.thestage_auth_token,
            slug=unique_id,
        )
        if not get_inference_simulator_response:
            typer.echo(__("Inference simulator with UID %uid% not found", {"uid": unique_id}))
            raise typer.Exit(1)
        else:
            inference_simulator_id = get_inference_simulator_response.inferenceSimulator.id
            logging_service.print_last_inference_simulator_logs(config=config, inference_simulator_id=inference_simulator_id, logs_number=logs_number)

    app_logger.info(f'Inference simulator logs - end')
    raise typer.Exit(0)


@inference_simulator_model_app.command("deploy-instance", no_args_is_help=True, help=__("Deploy an inference simulator model to an instance"))
def deploy_inference_simulator_model_to_instance(
        unique_id: Optional[str] = typer.Argument(help=__("Inference simulator model unique ID"), ),
        rented_instance_unique_id: Optional[str] = typer.Option(
            None,
            '--rented-instance-unique-id',
            '-ruid',
            help=__("The rented instance unique ID on which the inference simulator model will be deployed"),
            is_eager=False,
        ),
        self_hosted_instance_unique_id: Optional[str] = typer.Option(
            None,
            '--self-hosted-instance-unique-id',
            '-suid',
            help=__("The self-hosted instance unique ID on which the inference simulator model will be deployed"),
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory. By default, the current directory is used"),
            show_default=False,
            is_eager=False,
        ),
        enable_log_stream: Optional[bool] = typer.Option(
            True,
            " /--no-logs",
            " /-nl",
            help=__("Disable real-time log streaming"),
            is_eager=False,
        ),
):
    """
        Deploy an inference simulator model to an instance
    """

    if unique_id and not re.match(r"^[a-zA-Z0-9-]+$", unique_id):
        raise typer.BadParameter(__("Invalid UID format. The UID can only contain letters, numbers, and hyphens."))

    unique_id_with_timestamp = f"{unique_id}-{int(time.time())}"

    app_logger.info(f'Deploying an inference simulator model')

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    config = service_factory.get_config_provider().get_full_config()

    project_service = service_factory.get_project_service()

    project_service.project_deploy_inference_simulator_model_to_instance(
        config=config,
        unique_id=unique_id,
        unique_id_with_timestamp=unique_id_with_timestamp,
        rented_instance_unique_id=rented_instance_unique_id,
        self_hosted_instance_unique_id=self_hosted_instance_unique_id,
    )

    if enable_log_stream:
        logging_service: LoggingService = service_factory.get_logging_service()

        logging_service.stream_inference_simulator_logs_with_controls(
            config=config,
            slug=unique_id_with_timestamp
        )
    raise typer.Exit(0)


@inference_simulator_model_app.command("deploy-sagemaker", no_args_is_help=True, help=__("Deploy an inference simulator model to SageMaker"))
def deploy_inference_simulator_model_to_sagemaker(
        unique_id: Optional[str] = typer.Argument(help=__("Inference simulator model unique ID"), ),
        arn: Optional[str] = typer.Option(
            None,
            '--amazon-resource-name',
            '-arn',
            help=__("The Amazon Resource Name of the IAM Role to use, e.g., arn:aws:iam::{aws_account_id}:role/{role}"),
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory. By default, the current directory is used"),
            show_default=False,
            is_eager=False,
        ),
        instance_type: Optional[str] = typer.Option(
            None,
            '--instance-type',
            '-it',
            help=__("Instance type on which the inference simulator model will be deployed"),
            is_eager=False,
        ),
        initial_variant_weight: Optional[float] = typer.Option(
            None,
            "--initial-variant-weight",
            "-ivw",
            help=__("Initial Variant Weight. By default 1.0"),
            show_default=False,
            is_eager=False,
        ),
        initial_instance_count: Optional[int] = typer.Option(
            None,
            "--initial-instance-count",
            "-iic",
            help=__("Initial Instance Count"),
            show_default=False,
            is_eager=False,
        ),

):
    """
        Deploy an inference simulator model to SageMaker
    """

    if unique_id and not re.match(r"^[a-zA-Z0-9-]+$", unique_id):
        raise typer.BadParameter(__("Invalid UID format. The UID can only contain letters, numbers, and hyphens."))

    app_logger.info(f'Deploying an inference simulator model')

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    config = service_factory.get_config_provider().get_full_config()

    project_service = service_factory.get_project_service()

    project_service.project_deploy_inference_simulator_model_to_sagemaker(
        config=config,
        unique_id=unique_id,
        arn=arn,
        instance_type=instance_type,
        initial_variant_weight=initial_variant_weight,
        initial_instance_count=initial_instance_count,
    )



import os
import time
from datetime import datetime
from pathlib import Path
from typing import Optional, List

import json

import boto3
import click
import typer
from git import Commit
from tabulate import tabulate

from thestage.services.clients.thestage_api.core.http_client_exception import HttpClientException
from thestage.services.core_files.config_entity import ConfigEntity
from thestage.color_scheme.color_scheme import ColorScheme
from thestage.entities.enums.yes_no_response import YesOrNoResponse
from thestage.exceptions.git_access_exception import GitAccessException
from thestage.i18n.translation import __
from thestage.services.clients.git.git_client import GitLocalClient
from thestage.services.clients.thestage_api.dtos.container_response import DockerContainerDto
from thestage.services.clients.thestage_api.dtos.enums.container_status import DockerContainerStatus
from thestage.services.clients.thestage_api.dtos.inference_controller.deploy_inference_model_to_instance_response import \
    DeployInferenceModelToInstanceResponse
from thestage.services.clients.thestage_api.dtos.inference_controller.deploy_inference_model_to_sagemaker_response import \
    DeployInferenceModelToSagemakerResponse
from thestage.services.clients.thestage_api.dtos.inference_controller.get_inference_simulator_response import \
    GetInferenceSimulatorResponse
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList
from thestage.services.clients.thestage_api.dtos.project_controller.project_push_inference_simulator_model_response import \
    ProjectPushInferenceSimulatorModelResponse
from thestage.services.clients.thestage_api.dtos.project_controller.project_run_task_response import \
    ProjectRunTaskResponse
from thestage.services.clients.thestage_api.dtos.project_controller.project_start_inference_simulator_response import \
    ProjectStartInferenceSimulatorResponse
from thestage.services.clients.thestage_api.dtos.project_response import ProjectDto
from thestage.services.clients.thestage_api.dtos.task_controller.task_view_response import TaskViewResponse
from thestage.services.filesystem_service import FileSystemServiceCore
from thestage.services.project.dto.inference_simulator_dto import InferenceSimulatorDto
from thestage.services.project.dto.inference_simulator_model_dto import InferenceSimulatorModelDto
from thestage.services.task.dto.task_dto import TaskDto
from thestage.services.project.dto.project_config import ProjectConfig
from thestage.services.project.mapper.project_task_mapper import ProjectTaskMapper
from thestage.services.remote_server_service import RemoteServerService
from thestage.services.abstract_service import AbstractService
from thestage.helpers.error_handler import error_handler
from thestage.services.clients.thestage_api.api_client import TheStageApiClient
from thestage.services.config_provider.config_provider import ConfigProvider
from rich import print


class ProjectService(AbstractService):
    __thestage_api_client: TheStageApiClient = None

    def __init__(
            self,
            thestage_api_client: TheStageApiClient,
            config_provider: ConfigProvider,
            remote_server_service: RemoteServerService,
            file_system_service: FileSystemServiceCore,
            git_local_client: GitLocalClient,
    ):
        super(ProjectService, self).__init__(
            config_provider=config_provider
        )
        self.__thestage_api_client = thestage_api_client
        self.__remote_server_service = remote_server_service
        self.__file_system_service = file_system_service
        self.__git_local_client = git_local_client
        self.__project_task_mapper = ProjectTaskMapper()
        self.__config_provider = config_provider


    @error_handler()
    def init_project(
            self,
            config: ConfigEntity,
            project_slug: str,
    ):
        project: Optional[ProjectDto] = self.__thestage_api_client.get_project_by_slug(
            slug=project_slug,
            token=config.main.thestage_auth_token,
        )

        if not project:
            typer.echo('Project not found')
            raise typer.Exit(1)

        is_git_folder = self.__git_local_client.is_present_local_git(
            path=config.runtime.working_directory,
        )
        if is_git_folder:
            has_remote = self.__git_local_client.has_remote(
                path=config.runtime.working_directory,
            )
            if has_remote:
                typer.echo(__('You have local repo with remote, we can not work with this'))
                raise typer.Exit(1)

        if not project.git_repository_url:
            typer.echo(__('Sketch dont have git repository url'))
            raise typer.Exit(1)

        if project.last_commit_hash or project.last_commit_description:
            continue_with_non_empty_repo: YesOrNoResponse = typer.prompt(
                text=__('Remote repository is probably not empty: latest commit is "{commit_description}" (sha: {commit_hash})\nDo you wish to continue?').format(commit_description=project.last_commit_description, commit_hash=project.last_commit_hash),
                show_choices=True,
                default=YesOrNoResponse.YES.value,
                type=click.Choice([r.value for r in YesOrNoResponse]),
                show_default=True,
            )
            if continue_with_non_empty_repo == YesOrNoResponse.NO:
                typer.echo(__('Project init aborted'))
                raise typer.Exit(0)

        deploy_ssh_key = self.__thestage_api_client.get_project_deploy_ssh_key(
            slug=project.slug,
            token=config.main.thestage_auth_token
        )

        deploy_key_path = self.__config_provider.save_project_deploy_ssh_key(
            deploy_ssh_key=deploy_ssh_key,
            project_slug=project.slug,
            project_id=project.id,
        )

        if is_git_folder:
            has_changes = self.__git_local_client.has_changes_with_untracked(
                path=config.runtime.working_directory,
            )
            if has_changes:
                typer.echo(__('You local repo has changes and not empty, please create empty folder'))
                raise typer.Exit(1)
        else:
            repo = self.__git_local_client.init_repository(
                path=config.runtime.working_directory,
            )

        is_remote_added = self.__git_local_client.add_remote_to_repo(
            path=config.runtime.working_directory,
            remote_url=project.git_repository_url,
            remote_name=project.git_repository_name,
        )
        if not is_remote_added:
            typer.echo(__('We can not add remote, something wrong'))
            raise typer.Exit(2)

        self.__git_local_client.git_fetch(path=config.runtime.working_directory, deploy_key_path=deploy_key_path)

        self.__git_local_client.init_gitignore(path=config.runtime.working_directory)

        self.__git_local_client.git_add_all(repo_path=config.runtime.working_directory)

        project_config = ProjectConfig()
        project_config.id = project.id
        project_config.slug = project.slug
        project_config.git_repository_url = project.git_repository_url
        project_config.deploy_key_path = str(deploy_key_path)
        self.__config_provider.save_project_config(project_config=project_config)


    @error_handler()
    def clone_project(
            self,
            config: ConfigEntity,
            project_slug: str,
    ):
        project: Optional[ProjectDto] = self.__thestage_api_client.get_project_by_slug(
            slug=project_slug,
            token=config.main.thestage_auth_token,
        )

        if not project:
            typer.echo('Project not found')
            raise typer.Exit(1)

        if not self.__file_system_service.is_folder_empty(folder=config.runtime.working_directory, auto_create=True):
            typer.echo(__("Cannot clone: the folder is not empty"))
            raise typer.Exit(1)

        is_git_folder = self.__git_local_client.is_present_local_git(
            path=config.runtime.working_directory,
        )

        if is_git_folder:
            typer.echo(__('You have local repo, we can not work with this'))
            raise typer.Exit(1)

        if not project.git_repository_url:
            typer.echo(__("Unexpected Project error, missing Repository"))
            raise typer.Exit(1)

        deploy_ssh_key = self.__thestage_api_client.get_project_deploy_ssh_key(slug=project.slug, token=config.main.thestage_auth_token)
        deploy_key_path = self.__config_provider.save_project_deploy_ssh_key(deploy_ssh_key=deploy_ssh_key, project_slug=project.slug, project_id=project.id)

        try:
            self.__git_local_client.clone(
                url=project.git_repository_url,
                path=config.runtime.working_directory,
                deploy_key_path=deploy_key_path
            )
            self.__git_local_client.init_gitignore(path=config.runtime.working_directory)
        except GitAccessException as ex:
            typer.echo(ex.get_message())
            typer.echo(ex.get_dop_message())
            typer.echo(__(
                "Please check you mail or open this repo url %git_url% and 'Accept invitation'",
                {
                    'git_url': ex.get_url()
                }
            ))
            raise typer.Exit(1)

        project_config = ProjectConfig()
        project_config.id = project.id
        project_config.slug = project.slug
        project_config.git_repository_url = project.git_repository_url
        project_config.deploy_key_path = str(deploy_key_path)
        self.__config_provider.save_project_config(project_config=project_config)


    @error_handler()
    def project_run_task(
            self,
            config: ConfigEntity,
            run_command: str,
            task_title: Optional[str] = None,
            commit_hash: Optional[str] = None,
            docker_container_slug: Optional[str] = None,
    ) -> Optional[TaskDto]:
        project_config: ProjectConfig = self.__get_fixed_project_config(config=config)
        if not project_config:
            typer.echo(__("No project found at the path: %path%. Please initialize or clone a project first.", {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        if not docker_container_slug and not project_config.default_container_uid:
            typer.echo(__('Docker container unique ID is required'))
            raise typer.Exit(1)

        container_slug_for_task = docker_container_slug if docker_container_slug else project_config.default_container_uid

        if not docker_container_slug:
            typer.echo(f"Using default docker container for this project: '{container_slug_for_task}'")

        container: DockerContainerDto = self.__thestage_api_client.get_container(
            token=config.main.thestage_auth_token,
            container_slug=container_slug_for_task,
        )

        if container is None:
            typer.echo(f"Could not find container '{container_slug_for_task}'")
            if project_config.default_container_uid == container_slug_for_task:
                project_config.default_container_uid = None
                project_config.prompt_for_default_container = True
                self.__config_provider.save_project_config(project_config=project_config)
                typer.echo(f"Default container settings were reset")
            raise typer.Exit(1)

        if container.project_id != project_config.id:
            typer.echo(f"Provided container '{docker_container_slug}' is not related to project '{project_config.slug}'")
            raise typer.Exit(1)

        if (project_config.prompt_for_default_container is None or project_config.prompt_for_default_container) and docker_container_slug and (project_config.default_container_uid != docker_container_slug):
            set_default_container_slug: str = typer.prompt(
                text=f"Would you like to set '{docker_container_slug}' as a default container for this project installation?",
                show_choices=True,
                default=YesOrNoResponse.YES.value,
                type=click.Choice([r.value for r in YesOrNoResponse]),
                show_default=True,
            )
            project_config.prompt_for_default_container = False
            if set_default_container_slug == YesOrNoResponse.YES.value:
                project_config.default_container_uid = docker_container_slug

            self.__config_provider.save_project_config(project_config=project_config)

        if not commit_hash:
            is_git_folder = self.__git_local_client.is_present_local_git(path=config.runtime.working_directory)
            if not is_git_folder:
                typer.echo("Error: working directory does not contain git repository")
                raise typer.Exit(1)

            is_commit_allowed: bool = True
            has_changes = self.__git_local_client.has_changes_with_untracked(
                path=config.runtime.working_directory,
            )

            if self.__git_local_client.is_head_detached(path=config.runtime.working_directory):
                is_commit_allowed = False
                print(f"[{ColorScheme.GIT_HEADLESS.value}]HEAD is detached[{ColorScheme.GIT_HEADLESS.value}]")

                is_headless_commits_present = self.__git_local_client.is_head_committed_in_headless_state(path=config.runtime.working_directory)
                if is_headless_commits_present:
                    print(f"[{ColorScheme.GIT_HEADLESS.value}]Current commit was made in detached head state. Cannot use it to run the task. Consider using 'project checkout' command to return to a valid reference.[{ColorScheme.GIT_HEADLESS.value}]")
                    raise typer.Exit(1)

                if has_changes:
                    print(f"[{ColorScheme.GIT_HEADLESS.value}]Local changes detected in detached head state. They will not impact the task execution.[{ColorScheme.GIT_HEADLESS.value}]")
                    response: YesOrNoResponse = typer.prompt(
                        text=__('Continue?'),
                        show_choices=True,
                        default=YesOrNoResponse.YES.value,
                        type=click.Choice([r.value for r in YesOrNoResponse]),
                        show_default=True,
                    )
                    if response == YesOrNoResponse.NO:
                        raise typer.Exit(0)

            if is_commit_allowed:
                self.__git_local_client.git_add_all(repo_path=config.runtime.working_directory)

                if has_changes:
                    branch_name = self.__git_local_client.get_active_branch_name(config.runtime.working_directory)
                    diff_stat = self.__git_local_client.git_diff_stat(repo_path=config.runtime.working_directory)
                    typer.echo(__('Active branch [%branch_name%] has uncommitted changes: %diff_stat_bottomline%', {
                        'diff_stat_bottomline': diff_stat,
                        'branch_name': branch_name,
                    }))

                    response: str = typer.prompt(
                        text=__('Commit changes?'),
                        show_choices=True,
                        default=YesOrNoResponse.YES.value,
                        type=click.Choice([r.value for r in YesOrNoResponse]),
                        show_default=True,
                    )
                    if response == YesOrNoResponse.NO.value:
                        typer.echo("Task cannot use uncommitted changes - aborting")
                        raise typer.Exit(0)

                    commit_name = typer.prompt(
                        text=__('Please provide commit message'),
                        show_choices=False,
                        type=str,
                        show_default=False,
                    )

                    if commit_name:
                        commit_result = self.__git_local_client.commit_local_changes(
                            path=config.runtime.working_directory,
                            name=commit_name
                        )

                        if commit_result:
                            # in docs not Commit object, on real - str
                            if isinstance(commit_result, str):
                                typer.echo(commit_result)
                    else:
                        typer.echo(__('Cannot commit with empty commit message'))
                        raise typer.Exit(0)
                else:
                    pass
                    # possible to push new empty branch - only that there's a wrong place to do so

                self.__git_local_client.push_changes(
                    path=config.runtime.working_directory,
                    deploy_key_path=project_config.deploy_key_path
                )
                typer.echo(__("Pushed changes to remote repository"))

            commit = self.__git_local_client.get_current_commit(path=config.runtime.working_directory)
            if commit and isinstance(commit, Commit):
                commit_hash = commit.hexsha

        commit = self.__git_local_client.get_commit_by_hash(path=config.runtime.working_directory, commit_hash=commit_hash)
        if commit and isinstance(commit, Commit):
            if not task_title:
                task_title = commit.message.strip()
        else:
            print(f'[red]Error: commit \'{commit_hash}\' was not found in the local repository[/red]')
            raise typer.Exit(0)

        if not task_title:  # should not happen but maybe git allows some kind of empty messages
            task_title = f'Task_{commit_hash}'
            typer.echo(f'Commit message is empty. Task title is set to "{task_title}"')

        run_task_response: ProjectRunTaskResponse = self.__thestage_api_client.execute_project_task(
            token=config.main.thestage_auth_token,
            project_slug=project_config.slug,
            docker_container_slug=container_slug_for_task,
            run_command=run_command,
            commit_hash=commit_hash,
            task_title=task_title,
        )
        if run_task_response:
            if run_task_response.message:
                print(f"[{ColorScheme.WARNING.value}]{run_task_response.message}[{ColorScheme.WARNING.value}]")
            if run_task_response.is_success and run_task_response.task:
                typer.echo(f"Task '{run_task_response.task.title}' has been scheduled successfully. Task ID: {run_task_response.task.id}")
                return run_task_response.task
            else:
                typer.echo(f'The task failed with an error: {run_task_response.message}')
                raise typer.Exit(1)
        else:
            typer.echo("The task failed with an error")
            raise typer.Exit(1)

    @error_handler()
    def cancel_task(self, task_id: int, config: ConfigEntity):
        cancel_result = self.__thestage_api_client.cancel_task(
            token=config.main.thestage_auth_token,
            task_id=task_id,
        )

        if cancel_result.is_success:
            typer.echo(f'Task {task_id} has been canceled')
        else:
            typer.echo(f'Task {task_id} could not be canceled: {cancel_result.message}')


    @error_handler()
    def project_run_inference_simulator(
            self,
            config: ConfigEntity,
            slug: Optional[str] = None,
            commit_hash: Optional[str] = None,
            rented_instance_unique_id: Optional[str] = None,
            self_hosted_instance_unique_id: Optional[str] = None,
            inference_dir: Optional[str] = None,
            is_skip_installation: Optional[bool] = False,
    ) -> Optional[InferenceSimulatorDto]:
        project_config: ProjectConfig = self.__get_fixed_project_config(config=config)
        if not project_config:
            typer.echo(__("No project found at the path: %path%. Please initialize or clone a project first. Or provide path to project using --working-directory option.",
                          {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        if rented_instance_unique_id and self_hosted_instance_unique_id:
            typer.echo(__("Error: Cannot provide both rented and self-hosted instance unique IDs."))
            raise typer.Exit(1)

        if not rented_instance_unique_id and not self_hosted_instance_unique_id:
            typer.echo(__("Error: Either a rented instance ID or a self-hosted instance unique ID must be provided."))
            raise typer.Exit(1)

        project_config: ProjectConfig = self.__config_provider.read_project_config()
        if not project_config:
            typer.echo(__("No project found at the path: %path%. Please initialize or clone a project first.",
                          {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        if not commit_hash:
            is_git_folder = self.__git_local_client.is_present_local_git(path=config.runtime.working_directory)
            if not is_git_folder:
                typer.echo("Error: Working directory does not contain git repository.")
                raise typer.Exit(1)

            is_commit_allowed: bool = True
            has_changes = self.__git_local_client.has_changes_with_untracked(
                path=config.runtime.working_directory,
            )

            if self.__git_local_client.is_head_detached(path=config.runtime.working_directory):
                print(f"[{ColorScheme.GIT_HEADLESS.value}]HEAD is detached[{ColorScheme.GIT_HEADLESS.value}]")

                is_headless_commits_present = self.__git_local_client.is_head_committed_in_headless_state(
                    path=config.runtime.working_directory)
                if is_headless_commits_present:
                    print(
                        f"[{ColorScheme.GIT_HEADLESS.value}]Current commit was made in detached head state. Cannot use it to start the inference simulator. Consider using 'project checkout' command to return to a valid reference.[{ColorScheme.GIT_HEADLESS.value}]")
                    raise typer.Exit(1)

                if has_changes:
                    print(
                        f"[{ColorScheme.GIT_HEADLESS.value}]Local changes detected in detached head state. They will not impact the inference simulator.[{ColorScheme.GIT_HEADLESS.value}]")
                    is_commit_allowed = False
                    response: YesOrNoResponse = typer.prompt(
                        text=__('Continue?'),
                        show_choices=True,
                        default=YesOrNoResponse.YES.value,
                        type=click.Choice([r.value for r in YesOrNoResponse]),
                        show_default=True,
                    )
                    if response == YesOrNoResponse.NO:
                        raise typer.Exit(0)

            if is_commit_allowed:
                self.__git_local_client.git_add_all(repo_path=config.runtime.working_directory)

                if has_changes:
                    branch_name = self.__git_local_client.get_active_branch_name(config.runtime.working_directory)
                    diff_stat = self.__git_local_client.git_diff_stat(repo_path=config.runtime.working_directory)
                    typer.echo(__('Active branch [%branch_name%] has uncommitted changes: %diff_stat_bottomline%', {
                        'diff_stat_bottomline': diff_stat,
                        'branch_name': branch_name,
                    }))

                    response: str = typer.prompt(
                        text=__('Commit changes?'),
                        show_choices=True,
                        default=YesOrNoResponse.YES.value,
                        type=click.Choice([r.value for r in YesOrNoResponse]),
                        show_default=True,
                    )
                    if response == YesOrNoResponse.NO.value:
                        typer.echo("inference simulator cannot use uncommitted changes - aborting")
                        raise typer.Exit(0)

                    commit_name = typer.prompt(
                        text=__('Please provide commit message'),
                        show_choices=False,
                        type=str,
                        show_default=False,
                    )

                    if commit_name:
                        commit_result = self.__git_local_client.commit_local_changes(
                            path=config.runtime.working_directory,
                            name=commit_name
                        )

                        if commit_result:
                            # in docs not Commit object, on real - str
                            if isinstance(commit_result, str):
                                typer.echo(commit_result)

                        self.__git_local_client.push_changes(
                            path=config.runtime.working_directory,
                            deploy_key_path=project_config.deploy_key_path
                        )
                        typer.echo(__("Pushed changes to remote repository"))
                    else:
                        typer.echo(__('Cannot commit with empty commit name, your code will run without last changes.'))
                else:
                    pass
                    # possible to push new empty branch - only that there's a wrong place to do so
            commit = self.__git_local_client.get_current_commit(path=config.runtime.working_directory)
            if commit and isinstance(commit, Commit):
                commit_hash = commit.hexsha

        if not slug:
            typer.echo(__("Please provide inference simulator unique ID"))

            auto_title = "Inference simulator within project {slug}".format(slug=project_config.slug)
            uid: str = typer.prompt(
                text=__('New inference simulator unique ID:'),
                show_choices=False,
                default=auto_title,
                type=str,
                show_default=True,
            )
            if not uid:
                typer.echo(__("Inference simulator unique ID cannot be empty"))
                raise typer.Exit(1)
            else:
                slug = uid

        start_inference_simulator_response: ProjectStartInferenceSimulatorResponse = self.__thestage_api_client.start_project_inference_simulator(
            token=config.main.thestage_auth_token,
            project_slug=project_config.slug,
            commit_hash=commit_hash,
            slug=slug,
            rented_instance_unique_id=rented_instance_unique_id,
            self_hosted_instance_unique_id=self_hosted_instance_unique_id,
            inference_dir=inference_dir,
            is_skip_installation=is_skip_installation,
        )
        if start_inference_simulator_response:
            if start_inference_simulator_response.message:
                typer.echo(start_inference_simulator_response.message)
            if start_inference_simulator_response.is_success and start_inference_simulator_response.inferenceSimulator:
                typer.echo("Inference simulator has been scheduled to run successfully.")
                return start_inference_simulator_response.inferenceSimulator
            else:
                typer.echo(__(
                    'Inference simulator failed to run with an error: %server_massage%',
                    {'server_massage': start_inference_simulator_response.message or ""}
                ))
                raise typer.Exit(1)
        else:
            typer.echo(__("Inference simulator failed to run with an error"))
            raise typer.Exit(1)


    @error_handler()
    def project_push_inference_simulator(
            self,
            config: ConfigEntity,
            slug: Optional[str] = None,
    ):

        push_inference_simulator_model_response: ProjectPushInferenceSimulatorModelResponse = self.__thestage_api_client.push_project_inference_simulator_model(
            token=config.main.thestage_auth_token,
            slug=slug,
        )
        if push_inference_simulator_model_response:
            if push_inference_simulator_model_response.message:
                typer.echo(push_inference_simulator_model_response.message)
            if push_inference_simulator_model_response.is_success:
                typer.echo("Inference simulator has been successfully scheduled to be pushed to S3 and ECR.")
            else:
                typer.echo(__(
                    'Failed to push inference simulator with an error: %server_massage%',
                    {'server_massage': push_inference_simulator_model_response.message or ""}
                ))
                raise typer.Exit(1)
        else:
            typer.echo(__("Failed to push inference simulator with an error"))
            raise typer.Exit(1)

    @error_handler()
    def project_get_and_save_inference_simulator_metadata(
            self,
            config: ConfigEntity,
            slug: Optional[str] = None,
            file_path: Optional[str] = None,
    ):
        get_inference_metadata_response: GetInferenceSimulatorResponse = self.__thestage_api_client.get_inference_simulator(
            token=config.main.thestage_auth_token,
            slug=slug,
        )

        metadata = get_inference_metadata_response.inferenceSimulator.qlip_serve_metadata

        if metadata:
            typer.echo("qlip_serve_metadata:")
            typer.echo(json.dumps(json.loads(metadata), indent=4))

            if not file_path:
                file_path = Path(os.getcwd()) / "metadata.json"
                typer.echo(__("No file path provided. Saving metadata to %file_path%", {"file_path": str(file_path)}))

            try:
                parsed_metadata = json.loads(metadata)

                output_file = Path(file_path)
                output_file.parent.mkdir(parents=True, exist_ok=True)
                with output_file.open("w", encoding="utf-8") as file:
                    json.dump(parsed_metadata, file, indent=4)
                typer.echo(__("Metadata successfully saved to %file_path%", {"file_path": str(file_path)}))
            except Exception as e:
                typer.echo(__("Failed to save metadata to %file_path%. Error: %error%",
                              {"file_path": file_path, "error": str(e)}))
                raise typer.Exit(1)
        else:
            typer.echo(__("No qlip_serve_metadata found"))
            raise typer.Exit(1)


    @error_handler()
    def get_project_inference_simulator_list(
            self,
            config: ConfigEntity,
            project_slug: str,
            statuses: List[str],
            row: int = 5,
            page: int = 1,
    ) -> PaginatedEntityList[InferenceSimulatorDto]:
        data: Optional[PaginatedEntityList[InferenceSimulatorDto]] = self.__thestage_api_client.get_inference_simulator_list_for_project(
            token=config.main.thestage_auth_token,
            statuses=statuses,
            project_slug=project_slug,
            page=page,
            limit=row,
        )

        return data


    @error_handler()
    def get_project_inference_simulator_model_list(
            self,
            config: ConfigEntity,
            project_slug: str,
            statuses: List[str],
            row: int = 5,
            page: int = 1,
    ) -> PaginatedEntityList[InferenceSimulatorModelDto]:
        data: Optional[
            PaginatedEntityList[InferenceSimulatorModelDto]] = self.__thestage_api_client.get_inference_simulator_model_list_for_project(
            token=config.main.thestage_auth_token,
            statuses=statuses,
            project_slug=project_slug,
            page=page,
            limit=row,
        )

        return data


    @error_handler()
    def get_project_task_list(
            self,
            config: ConfigEntity,
            project_slug: str,
            row: int = 5,
            page: int = 1,
    ) -> PaginatedEntityList[TaskDto]:
        data: Optional[PaginatedEntityList[TaskDto]] = self.__thestage_api_client.get_task_list_for_project(
            token=config.main.thestage_auth_token,
            project_slug=project_slug,
            page=page,
            limit=row,
        )

        return data


    @error_handler()
    def checkout_project(
            self,
            config: ConfigEntity,
            task_id: Optional[int],
            branch_name: Optional[str],
    ):
        project_config: ProjectConfig = self.__get_fixed_project_config(config=config)
        if not project_config:
            typer.echo(__("This command is only allowed from within an initialized project directory"))
            raise typer.Exit(1)

        target_commit_hash: Optional[str] = None
        if task_id:
            task_view_response: Optional[TaskViewResponse] = None
            try:
                task_view_response = self.__thestage_api_client.get_task(
                    token=config.main.thestage_auth_token,
                    task_id=task_id,
                )
            except HttpClientException as e:
                if e.get_status_code() == 400:
                    typer.echo(f"Task {task_id} was not found")
                    # overriding arguments here
                    branch_name = str(task_id)
                    task_id = None

            if task_view_response and task_view_response.task:
                target_commit_hash = task_view_response.task.commit_hash
                if not target_commit_hash:
                    typer.echo(f"Provided task ({task_id}) has no commit hash")  # possible legacy problems
                    raise typer.Exit(1)

        is_commit_allowed: bool = True

        if self.__git_local_client.is_head_detached(path=config.runtime.working_directory):
            is_commit_allowed = False
            if self.__git_local_client.is_head_committed_in_headless_state(path=config.runtime.working_directory):
                commit_message = self.__git_local_client.get_current_commit(path=config.runtime.working_directory).message
                print(f"[{ColorScheme.GIT_HEADLESS.value}]Your current commit '{commit_message.strip()}' was likely created in detached head state. Checking out will discard all changes.[/{ColorScheme.GIT_HEADLESS.value}]")
                response: YesOrNoResponse = typer.prompt(
                    text=__('Continue?'),
                    show_choices=True,
                    default=YesOrNoResponse.YES.value,
                    type=click.Choice([r.value for r in YesOrNoResponse]),
                    show_default=True,
                )
                if response == YesOrNoResponse.NO:
                    raise typer.Exit(0)
        else:
            if self.__git_local_client.get_active_branch_name(path=config.runtime.working_directory) == branch_name:
                typer.echo(f"You are already at branch '{branch_name}'")
                raise typer.Exit(0)

        if is_commit_allowed:
            self.__git_local_client.git_add_all(repo_path=config.runtime.working_directory)

            has_changes = self.__git_local_client.has_changes_with_untracked(
                path=config.runtime.working_directory,
            )
            if has_changes:
                active_branch_name = self.__git_local_client.get_active_branch_name(config.runtime.working_directory)
                diff_stat = self.__git_local_client.git_diff_stat(repo_path=config.runtime.working_directory)
                typer.echo(__('Active branch [%branch_name%] has uncommitted changes: %diff_stat_bottomline%', {
                    'diff_stat_bottomline': diff_stat,
                    'branch_name': active_branch_name,
                }))

                response: str = typer.prompt(
                    text=__('Commit changes?'),
                    show_choices=True,
                    default=YesOrNoResponse.YES.value,
                    type=click.Choice([r.value for r in YesOrNoResponse]),
                    show_default=True,
                )
                if response == YesOrNoResponse.NO.value:
                    typer.echo(__('Cannot checkout with uncommitted changes'))
                    raise typer.Exit(0)

                commit_name = typer.prompt(
                    text=__('Please provide commit message'),
                    show_choices=False,
                    type=str,
                    show_default=False,
                )

                if commit_name:
                    commit_result = self.__git_local_client.commit_local_changes(
                        path=config.runtime.working_directory,
                        name=commit_name
                    )

                    if commit_result:
                        # in docs not Commit object, on real - str
                        if isinstance(commit_result, str):
                            typer.echo(commit_result)

                    self.__git_local_client.push_changes(
                        path=config.runtime.working_directory,
                        deploy_key_path=project_config.deploy_key_path
                    )
                    typer.echo(__("Pushed changes to remote repository"))
                else:
                    typer.echo(__('Cannot commit with empty commit name'))
                    raise typer.Exit(0)

        if target_commit_hash:
            if self.__git_local_client.get_current_commit(path=config.runtime.working_directory).hexsha != target_commit_hash:
                is_checkout_successful = self.__git_local_client.git_checkout_to_commit(
                    path=config.runtime.working_directory,
                    commit_hash=target_commit_hash
                )

                if is_checkout_successful:
                    print(f"Checked out to commit {target_commit_hash}")
                    print(f"[{ColorScheme.GIT_HEADLESS.value}]HEAD is detached. To be able make changes in repository, checkout to any branch.[/{ColorScheme.GIT_HEADLESS.value}]")
            else:
                typer.echo("HEAD is already at requested commit")
        elif branch_name:
            if self.__git_local_client.is_branch_exists(path=config.runtime.working_directory, branch_name=branch_name):
                self.__git_local_client.git_checkout_to_branch(
                    path=config.runtime.working_directory,
                    branch=branch_name
                )
                typer.echo(f"Checked out to branch '{branch_name}'")
            else:
                typer.echo(f"Branch '{branch_name}' was not found in project repository")
        else:
            main_branch = self.__git_local_client.find_main_branch_name(path=config.runtime.working_directory)
            if main_branch:
                self.__git_local_client.git_checkout_to_branch(
                    path=config.runtime.working_directory,
                    branch=main_branch
                )
                typer.echo(f"Checked out to detected main branch: '{main_branch}'")
            else:
                typer.echo("No main branch found")




    @error_handler()
    def set_default_container(self, config: ConfigEntity, container_uid: Optional[str]):
        project_config: ProjectConfig = self.__config_provider.read_project_config()

        if project_config is None:
            typer.echo(f"No project found in working directory")
            raise typer.Exit(1)

        if container_uid:
            container: DockerContainerDto = self.__thestage_api_client.get_container(
                token=config.main.thestage_auth_token,
                container_slug=container_uid,
            )
            if container is None:
                typer.echo(f"Could not find container '{container_uid}'")
                raise typer.Exit(1)

            if container.project_id != project_config.id:
                typer.echo(f"Provided container '{container_uid}' is not related to project '{project_config.slug}'")
                raise typer.Exit(1)

            if container.frontend_status.status_key != DockerContainerStatus.RUNNING:
                typer.echo(f"Note: provided container '{container_uid}' is in status '{container.frontend_status.status_translation}'")

        project_config.default_container_uid = container_uid
        project_config.prompt_for_default_container = False
        self.__config_provider.save_project_config(project_config=project_config)
        typer.echo("Default container settings were updated")


    @error_handler()
    def unset_default_container(self):
        project_config: ProjectConfig = self.__config_provider.read_project_config()

        if project_config is None:
            typer.echo(f"No project found in working directory")
            raise typer.Exit(1)

        project_config.default_container_uid = None
        project_config.prompt_for_default_container = True  # True or False?
        self.__config_provider.save_project_config(project_config=project_config)
        typer.echo("Default container settings were updated")


    @error_handler()
    def print_project_config(self, config):
        project_config: ProjectConfig = self.__config_provider.read_project_config()

        if project_config is None:
            typer.echo(f"No project found in working directory")
            raise typer.Exit(1)

        is_deploy_key_exists = project_config.deploy_key_path and self.__file_system_service.check_if_path_exist(project_config.deploy_key_path)

        typer.echo(tabulate(
            [
                [
                    "Project unique ID", project_config.slug
                ],
                [
                    "Default docker container unique ID", project_config.default_container_uid if project_config.default_container_uid else "<None>"
                ],
                [
                    "Deploy key path", project_config.deploy_key_path if is_deploy_key_exists else "<None>"
                ],
            ],
            showindex=False,
            tablefmt="simple",
        ))

        if is_deploy_key_exists:
            typer.echo("")
            typer.echo(f"You can insert the following text:")
            print(f"[{ColorScheme.USEFUL_INFO.value}]GIT_SSH_COMMAND=\"ssh -o StrictHostKeyChecking=no -o IdentitiesOnly=yes -i {project_config.deploy_key_path}\"[{ColorScheme.USEFUL_INFO.value}]")
            typer.echo(f"before any regular git command to manage your local Project repository directly")

    @error_handler()
    def __get_fixed_project_config(self, config: ConfigEntity) -> Optional[ProjectConfig]:
        project_config: ProjectConfig = self.__config_provider.read_project_config()
        if project_config is None:
            return None

        if not Path(project_config.deploy_key_path).is_file():
            deploy_ssh_key = self.__thestage_api_client.get_project_deploy_ssh_key(
                slug=project_config.slug,
                token=config.main.thestage_auth_token
            )

            deploy_key_path = self.__config_provider.save_project_deploy_ssh_key(
                deploy_ssh_key=deploy_ssh_key,
                project_slug=project_config.slug,
                project_id=project_config.id,
            )

            project_config.deploy_key_path = deploy_key_path
            self.__config_provider.save_project_config(project_config=project_config)
            typer.echo(f'Recreated missing deploy key for the project')

        return project_config

    @error_handler()
    def project_deploy_inference_simulator_model_to_instance(
            self,
            config: ConfigEntity,
            unique_id: Optional[str] = None,
            unique_id_with_timestamp: Optional[str] = None,
            rented_instance_unique_id: Optional[str] = None,
            self_hosted_instance_unique_id: Optional[str] = None,
    ) -> None:
        project_config: ProjectConfig = self.__get_fixed_project_config(config=config)
        if not project_config:
            typer.echo(
                __("No project found at the path: %path%. Please initialize or clone a project first. Or provide path to project using --working-directory option.",
                   {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        if rented_instance_unique_id and self_hosted_instance_unique_id:
            typer.echo(__("Error: Cannot provide both rented and self-hosted instance unique IDs."))
            raise typer.Exit(1)

        if not rented_instance_unique_id and not self_hosted_instance_unique_id:
            typer.echo(__("Error: Either a rented instance ID or a self-hosted instance unique ID must be provided."))
            raise typer.Exit(1)

        project_config: ProjectConfig = self.__config_provider.read_project_config()
        if not project_config:
            typer.echo(__("No project found at the path: %path%. Please initialize or clone a project first.",
                          {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        typer.echo(__("Creating inference simulator with unique ID: %unique_id_with_timestamp%", {"unique_id_with_timestamp": unique_id_with_timestamp}))
        deploy_model_to_instance_response: DeployInferenceModelToInstanceResponse = self.__thestage_api_client.deploy_inference_model_to_instance(
            token=config.main.thestage_auth_token,
            unique_id=unique_id,
            unique_id_with_timestamp=unique_id_with_timestamp,
            rented_instance_unique_id=rented_instance_unique_id,
            self_hosted_instance_unique_id=self_hosted_instance_unique_id
        )
        if deploy_model_to_instance_response:
            if deploy_model_to_instance_response.message:
                typer.echo(deploy_model_to_instance_response.message)
            if deploy_model_to_instance_response.is_success:
                typer.echo("Inference simulator has been scheduled to run successfully.")
            else:
                typer.echo(__(
                    'Inference simulator failed to run with an error: %server_massage%',
                    {'server_massage': deploy_model_to_instance_response.message or ""}
                ))
                raise typer.Exit(1)
        else:
            typer.echo(__("Inference simulator failed to run with an error"))
            raise typer.Exit(1)


    @error_handler()
    def project_deploy_inference_simulator_model_to_sagemaker(
            self,
            config: ConfigEntity,
            unique_id: Optional[str] = None,
            arn: Optional[str] = None,
            instance_type: Optional[str] = None,
            initial_variant_weight: Optional[float] = 1.0,
            initial_instance_count: Optional[int] = None,
    ) -> None:
        project_config: ProjectConfig = self.__get_fixed_project_config(config=config)
        if not project_config:
            typer.echo(
                __("No project found at the path: %path%. Please initialize or clone a project first. Or provide path to project using --working-directory option.",
                   {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        if not instance_type:
            typer.echo(__("Error: Instance type is required."))
            raise typer.Exit(1)

        if not initial_instance_count:
            typer.echo(__("Error: Initial instance count is required."))
            raise typer.Exit(1)

        if not arn:
            typer.echo(__("Error: ARN is required."))
            raise typer.Exit(1)

        project_config: ProjectConfig = self.__config_provider.read_project_config()
        if not project_config:
            typer.echo(__("No project found at the path: %path%. Please initialize or clone a project first.",
                          {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        deploy_model_to_sagemaker_response: DeployInferenceModelToSagemakerResponse = self.__thestage_api_client.deploy_inference_model_to_sagemaker(
            token=config.main.thestage_auth_token,
            unique_id=unique_id,
            arn=arn,
        )

        if not deploy_model_to_sagemaker_response.is_success:
            typer.echo(__(
                'Failed to prepare model for deployment with an error: %server_massage%',
                {'server_massage': deploy_model_to_sagemaker_response.message or ""}
            ))
            raise typer.Exit(1)

        model_id = deploy_model_to_sagemaker_response.modelId
        image_uri = deploy_model_to_sagemaker_response.ecrImageUrl
        model_uri = deploy_model_to_sagemaker_response.s3ArtifactsUrl
        region = "us-east-1"
        sm_client = boto3.client('sagemaker', region_name=region)

        try:
            container = {
                "Image": image_uri,
                "ModelDataUrl": model_uri,
                "Environment": {
                    "SAGEMAKER_TRITON_DEFAULT_MODEL_NAME": model_id,
                    "THESTAGE_API_URL": config.main.thestage_api_url,
                    "THESTAGE_AUTH_TOKEN": config.main.thestage_auth_token
                },
            }

            sm_model_name = f"{unique_id}-{datetime.now().strftime('%Y%m%d-%H%M%S')}"
            create_model_response = sm_client.create_model(
                ModelName=sm_model_name,
                ExecutionRoleArn=arn,
                PrimaryContainer=container,
            )
            typer.echo(f"Model created successfully. Model ARN: {create_model_response['ModelArn']}")

            endpoint_config_name = f"{unique_id}-{datetime.now().strftime('%Y%m%d-%H%M%S')}"
            create_endpoint_config_response = sm_client.create_endpoint_config(
                EndpointConfigName=endpoint_config_name,
                ProductionVariants=[
                    {
                        "InstanceType": instance_type,
                        "InitialVariantWeight": initial_variant_weight,
                        "InitialInstanceCount": initial_instance_count,
                        "ModelName": sm_model_name,
                        "VariantName": "AllTraffic",
                    }
                ],
            )
            typer.echo(
                f"Endpoint configuration created successfully. Endpoint Config ARN: {create_endpoint_config_response['EndpointConfigArn']}")

            endpoint_name = f"{unique_id}-{datetime.now().strftime('%Y%m%d-%H%M%S')}"
            create_endpoint_response = sm_client.create_endpoint(
                EndpointName=endpoint_name,
                EndpointConfigName=endpoint_config_name,
            )
            typer.echo(f"Endpoint created successfully. Endpoint ARN: {create_endpoint_response['EndpointArn']}")

            typer.echo("Waiting for the endpoint to become active...")
            while True:
                resp = sm_client.describe_endpoint(EndpointName=endpoint_name)
                status = resp["EndpointStatus"]
                typer.echo(f"Status: {status}")
                if status == "InService":
                    break
                elif status == "Failed":
                    typer.echo(f"Endpoint creation failed. Reason: {resp.get('FailureReason', 'Unknown')}")
                    raise typer.Exit(1)
                time.sleep(60)

            typer.echo(f"Endpoint is ready. ARN: {resp['EndpointArn']} Status: {status}")

        except Exception as e:
            typer.echo(__("Failed to deploy the inference simulator model to SageMaker: %error%", {"error": str(e)}))
            raise typer.Exit(1)


    @error_handler()
    def pull_project(self, config: ConfigEntity):
        project_config: ProjectConfig = self.__get_fixed_project_config(config=config)
        if not project_config:
            typer.echo(__("No project found at the path: %path%. Please initialize or clone a project first.", {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        typer.echo("Pulling code from remote repository...")
        self.__git_local_client.git_pull(
            path=config.runtime.working_directory,
            deploy_key_path=project_config.deploy_key_path,
        )


    @error_handler()
    def reset_project(self, config: ConfigEntity):
        project_config: ProjectConfig = self.__get_fixed_project_config(config=config)
        if not project_config:
            typer.echo(__("No project found at the path: %path%. Please initialize or clone a project first.", {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        typer.echo("Fetching code from remote repository...")
        self.__git_local_client.git_fetch(
            path=config.runtime.working_directory,
            deploy_key_path=project_config.deploy_key_path,
        )
        typer.echo("Resetting local branch...")
        self.__git_local_client.reset_hard(
            path=config.runtime.working_directory,
            deploy_key_path=project_config.deploy_key_path,
            reset_to_origin=True
        )



from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class RentedInstanceEntity(BaseModel):

    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True
    )

    # id: Optional[int] = Field(None, alias='ID')
    status: Optional[str] = Field(None, alias='STATUS')
    title: Optional[str] = Field(None, alias='TITLE')
    slug: Optional[str] = Field(None, alias='UNIQUE ID')
    cpu_type: Optional[str] = Field(None, alias='CPU TYPE')
    cpu_cores: Optional[str] = Field(None, alias='CPU CORES')
    gpu_type: Optional[str] = Field(None, alias='GPU TYPE')
    # ram_size_gb: Optional[int] = Field(None, alias='RAM SIZE GB')
    ip_address: Optional[str] = Field(None, alias='IP ADDRESS')
    # username: Optional[str] = Field(None, alias='USERNAME')
    # created_at: Optional[str] = Field(None, alias='CREATED AT')
    # updated_at: Optional[str] = Field(None, alias='UPDATED AT')



from typing import Optional
from pydantic import BaseModel, ConfigDict, Field

class InferenceSimulatorModelDto(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    id: Optional[int] = Field(None, alias='id')
    slug: Optional[str] = Field(None, alias='slug')
    client_id: Optional[int] = Field(None, alias='clientId')
    instance_rented_id: Optional[int] = Field(None, alias='instanceRentedId')
    selfhosted_instance_id: Optional[int] = Field(None, alias='selfhostedInstanceId')
    project_id: Optional[int] = Field(None, alias='projectId')
    status: Optional[str] = Field(None, alias='status')
    environment_metadata: Optional[str] = Field(None, alias='environmentMetadata')
    commit_hash: Optional[str] = Field(None, alias='commitHash')
    cpu_architecture: Optional[str] = Field(None, alias='cpuArchitecture')
    gpu_model: Optional[str] = Field(None, alias='gpuModel')
    ecr_image_url: Optional[str] = Field(None, alias='ecrImageUrl')
    s3_artifacts_url: Optional[str] = Field(None, alias='s3ArtifactsUrl')
    created_at: Optional[str] = Field(None, alias='createdAt')
    updated_at: Optional[str] = Field(None, alias='updatedAt')


from typing import Optional, List, Tuple

from thestage.services.clients.thestage_api.core.api_client_abstract import TheStageApiClientAbstract
from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage import __version__


class TheStageApiClientCore(TheStageApiClientAbstract):

    def __init__(self, timeout: int = 90, url: Optional[str] = None):
        super(TheStageApiClientCore, self).__init__(timeout=timeout, url=url)

    def validate_token(self, token: str) -> bool:
        data = {
            "userApiToken": token,
            "cliVersion": __version__,
        }

        response = self._request(
            method='POST',
            url='/user-api/v1/validate-token',
            data=data,
        )
        result = TheStageBaseResponse.model_validate(response) if response else None
        return result.is_success if result else False



from pathlib import Path
from typing import Optional

from thestage.i18n.translation import __
from thestage.helpers.logger.app_logger import app_logger
from thestage.controllers.utils_controller import get_current_directory, validate_config_and_get_service_factory
from thestage import __app_name__, __version__

import typer

from thestage.services.connect.connect_service import ConnectService

app = typer.Typer(no_args_is_help=True)


@app.command(no_args_is_help=False)
def version():
    """
        Returns the application's name and version
    """
    app_logger.info(f'Start version from {get_current_directory()}')
    typer.echo(
        __("%app_name% v%version%", {'app_name': __app_name__, 'version': __version__}))
    raise typer.Exit(0)


@app.command(name="connect", no_args_is_help=True, help=__("Connect to server instance or container using unique ID"))
def connect(
        uid: Optional[str] = typer.Argument(
            help=__("Unique ID of server instance or container"), ),
        username: Optional[str] = typer.Option(
            None,
            '--username',
            '-u',
            help=__("Username for the server instance (required when connecting to self-hosted instance)"),
            is_eager=False,
        ),
        private_ssh_key_path: str = typer.Option(
            None,
            "--private-key-path",
            "-pk",
            help=__("Path to private key that will be accepted by remote server (optional)"),
            is_eager=False,
        ),
):
    """
        Connects to entity with a unique ID
    """
    app_logger.info(f'Connect to some entity with UID')

    if private_ssh_key_path and not Path(private_ssh_key_path).is_file():
        typer.echo(f'No file found at provided path {private_ssh_key_path}')
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()

    connect_service: ConnectService = service_factory.get_connect_service()

    connect_service.connect_to_entity(uid=uid, username=username, private_key_path=private_ssh_key_path)


    app_logger.info(f'Stop connect to entity')
    raise typer.Exit(0)



import os
import shutil
from pathlib import Path
from typing import Optional, List

from thestage.entities.file_item import FileItemEntity
from thestage.exceptions.file_system_exception import FileSystemException


class FileSystemServiceCore:

    def get_ssh_path(self) -> Optional[Path]:
        home_path = self.get_home_path()
        ssh_path = home_path.joinpath('.ssh')
        if not ssh_path.exists():
            raise FileSystemException(f"Path does not exist: {ssh_path}")
        return ssh_path

    def get_home_path(self) -> Optional[Path]:
        try:
            return Path.home()
        except RuntimeError | OSError as ex1:
            raise FileSystemException("Error getting user home path") from ex1

    def create_if_not_exists_dir(self, path: Path) -> Path:
        if not path.exists():
            try:
                path.mkdir(exist_ok=True, parents=True)
            except FileNotFoundError as ex1:
                raise FileSystemException(message=f"FileNotFoundError (dir): {path}") from ex1
            except OSError as ex2:
                raise FileSystemException(message=f"Could not create directory: {path}") from ex2
        return path

    def create_if_not_exists_file(self, path: Path) -> Path:
        if not path.exists():
            try:
                path.touch(exist_ok=True)
            except FileNotFoundError as ex1:
                raise FileSystemException(message=f"FileNotFoundError (file): {path}") from ex1
            except OSError as ex2:
                raise FileSystemException(message=f"Could not create file: {path}") from ex2
        return path

    def get_path(self, directory: str, auto_create: bool = True) -> Path:
        path = Path(directory)
        if auto_create:
            self.create_if_not_exists_dir(path)
        return path

    def is_folder_empty(self, folder: str, auto_create: bool) -> bool:
        path = self.get_path(folder, auto_create)
        if not path.exists():
            return True
        if not path.is_dir():
            raise FileSystemException(message=f"Expected directory but found a file: {path}")
        objects = os.listdir(path)
        if len(objects) == 0:
            return True
        else:
            return False

    def is_folder_exists(self, folder: str, auto_create: bool = True) -> bool:
        path = self.get_path(folder, auto_create=auto_create)
        if path.exists():
            return True
        else:
            return False

    def find_line_in_text_file(self, file: str, find: str) -> bool:
        path = self.get_path(file, auto_create=False)
        if path and path.exists():
            with open(path, 'r') as file:
                for line in file.readlines():
                    if (find + "\n") == line:
                        return True
        return False

    def add_line_to_text_file(self, file: str, new_line: str):
        path = self.get_path(file, auto_create=False)
        if path and path.exists():
            with open(path, 'a') as file:
                file.write(new_line)
                file.write('\n')

    # TODO remove this fucking useless shit
    def check_if_path_exist(self, file: str) -> bool:
        path = self.get_path(file, auto_create=False)
        if path.exists():
            return True
        else:
            return False

    def get_path_items(self, folder: str) -> List[FileItemEntity]:
        path = self.get_path(folder, auto_create=False)
        path_items = []
        if not path.exists():
            return path_items

        parent = FileItemEntity.build_from_path(path=path)
        path_items.append(parent)
        if path.is_dir():
            objects = os.listdir(path)
            if objects:
                for item in objects:
                    elem = path.joinpath(item)
                    if elem.is_dir():
                        parent.children.extend(self.get_path_items(folder=str(elem)))
                    else:
                        parent.children.append(FileItemEntity.build_from_path(path=elem))
        return path_items

    def remove_folder(self, path: str):
        real_path = self.get_path(directory=path, auto_create=False)
        if real_path and real_path.exists():
            shutil.rmtree(real_path)
